from typing import Optional

from rdkit.Chem import FilterCatalog, Mol
from rdkit.Chem.rdfiltercatalog import FilterCatalogParams

from skfp.bases.base_filter import BaseFilter


class SureChEMBLFilter(BaseFilter):
    """
    SureChEMBL filter.

    Based on structural alerts, i.e. toxicophores. Filters out compounds likely to be
    toxic. Used at SureChEMBL, and based on the ToxAlerts publication [1]_, so many
    alerts are also found in OChem.

    The original page with rule definitions is available only by the Wayback Machine [2]_.
    They are also available in the RDKit code [3]_.

    Parameters
    ----------
    allow_one_violation : bool, default=False
        Whether to allow violating one of the rules for a molecule. This makes the
        filter less restrictive.

    return_indicators : bool, default=False
        Whether to return a binary vector with indicators which molecules pass the
        filter, instead of list of molecules.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform_x_y` and
        :meth:`transform` are parallelized over the input molecules. ``None`` means 1
        unless in a :obj:`joblib.parallel_backend` context. ``-1`` means using all
        processors. See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when filtering molecules.

    References
    ----------
    .. [1] `Iurii Sushko et al.
        "ToxAlerts: A Web Server of Structural Alerts for Toxic Chemicals and Compounds
        with Potential Adverse Reactions"
        J. Chem. Inf. Model. 2012, 52, 8, 2310–2316
        <https://pubs.acs.org/doi/10.1021/ci300245q>`_

    .. [2] `Wayback Machine - ChEMBL "Non MedChem-Friendly SMARTS"
        <https://web.archive.org/web/20230111133819/https://www.surechembl.org/knowledgebase/169485-non-medchem-friendly-smarts>`_

    .. [3] `RDKit SureChEMBL filter definitions
        <https://github.com/rdkit/rdkit/blob/e4f4644a89d6446ddebda0bf396fa4335324c41c/Code/GraphMol/FilterCatalog/chembl_surechembl.in>`_

    Examples
    --------
    >>> from skfp.filters import SureChEMBLFilter
    >>> smiles = ["C", "O", "C#CC#C"]
    >>> filt = SureChEMBLFilter()
    >>> filt
    SureChEMBLFilter()

    >>> filtered_mols = filt.transform(smiles)
    >>> filtered_mols
    ['C', 'O']
    """

    def __init__(
        self,
        allow_one_violation: bool = False,
        return_indicators: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            allow_one_violation=allow_one_violation,
            return_indicators=return_indicators,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )
        self._filters = self._load_filters()

    def _load_filters(self) -> FilterCatalog:
        filter_rules = FilterCatalogParams.FilterCatalogs.CHEMBL_SureChEMBL
        params = FilterCatalog.FilterCatalogParams()
        params.AddCatalog(filter_rules)
        filters = FilterCatalog.FilterCatalog(params)
        return filters

    def _apply_mol_filter(self, mol: Mol) -> bool:
        errors = len(self._filters.GetMatches(mol))
        return not errors or (self.allow_one_violation and errors == 1)



from abc import ABC
from copy import deepcopy
from numbers import Integral
from typing import Optional

from joblib import effective_n_jobs
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils._param_validation import InvalidParameterError
from tqdm import tqdm

from skfp.utils import run_in_parallel


class BasePreprocessor(ABC, BaseEstimator, TransformerMixin):
    """Base class for molecule preprocessing classes."""

    """Base class for fingerprint transformers."""

    # parameters common for all fingerprints
    _parameter_constraints: dict = {
        "n_jobs": [Integral, None],
        "batch_size": [Integral, None],
        "suppress_warnings": ["boolean"],
        "verbose": ["verbose"],
    }

    def __init__(
        self,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        suppress_warnings: bool = False,
        verbose: int = 0,
    ):
        self.n_jobs = n_jobs
        self.batch_size = batch_size
        self.suppress_warnings = suppress_warnings
        self.verbose = verbose

    def __sklearn_is_fitted__(self) -> bool:
        return True  # molecule preprocessing transformers don't need fitting

    def fit(self, X, y=None, **fit_params):
        """Unused, kept for Scikit-learn compatibility.

        Parameters
        ----------
        X : any
            Unused, kept for Scikit-learn compatibility.

        y : any
            Unused, kept for Scikit-learn compatibility.

        **fit_params : dict
            Unused, kept for Scikit-learn compatibility.

        Returns
        --------
        self
        """
        self._validate_params()
        return self

    def fit_transform(self, X, y=None, **fit_params):
        """
        The same as ``.transform()`` method, kept for Scikit-learn compatibility.

        Parameters
        ----------
        X : any
            See ``.transform()`` method.

        y : any
            See ``.transform()`` method.

        **fit_params : dict
            Unused, kept for Scikit-learn compatibility.

        Returns
        -------
        X_new : any
            See ``.transform()`` method.
        """
        return self.transform(X)

    def transform(self, X, copy: bool = False):
        self._validate_params()

        if copy:
            X = deepcopy(X)

        n_jobs = effective_n_jobs(self.n_jobs)
        if n_jobs == 1:
            if self.verbose:
                results = [self._transform_batch([mol]) for mol in tqdm(X)]
            else:
                results = self._transform_batch(X)
        else:
            results = run_in_parallel(
                self._transform_batch,
                data=X,
                n_jobs=n_jobs,
                batch_size=self.batch_size,
                flatten_results=True,
                verbose=self.verbose,
            )

        return results

    def _transform_batch(self, X):
        raise NotImplementedError

    def _validate_params(self) -> None:
        # override Scikit-learn validation to make stacktrace nicer
        try:
            super()._validate_params()
        except InvalidParameterError as e:
            raise InvalidParameterError(str(e)) from None



"""Classes for efficient hyperparameter optimization of fingerprint-based models."""

from .hyperparam_search import (
    FingerprintEstimatorGridSearch,
    FingerprintEstimatorRandomizedSearch,
)
from .splitters import (
    butina_train_test_split,
    butina_train_valid_test_split,
    maxmin_train_test_split,
    maxmin_train_valid_test_split,
    randomized_scaffold_train_test_split,
    randomized_scaffold_train_valid_test_split,
    scaffold_train_test_split,
    scaffold_train_valid_test_split,
)



from .aminoseq import MolFromAminoseqTransformer
from .inchi import MolFromInchiTransformer, MolToInchiTransformer
from .sdf import MolFromSDFTransformer, MolToSDFTransformer
from .smiles import MolFromSmilesTransformer, MolToSmilesTransformer



from collections.abc import Sequence
from numbers import Integral
from time import time
from typing import Optional, Union

import joblib
import numpy as np
from rdkit.Chem import Mol
from scipy.sparse import csr_array
from sklearn.base import BaseEstimator, _fit_context, clone
from sklearn.model_selection._search import BaseSearchCV, ParameterSampler
from sklearn.utils._param_validation import Interval, InvalidParameterError

from skfp.bases import BaseFingerprintTransformer


class FingerprintEstimatorRandomizedSearch(BaseEstimator):
    """
    Randomized search over specified hyperparameter distributions for a pipeline of
    a molecular fingerprint and scikit-learn estimator.

    This approach is useful for pipelines which first compute fingerprints and then
    operate on the resulting matrices, and when both fingerprint and estimator
    hyperparameters are optimized. Regular scikit-learn combination of ``Pipeline`` and
    ``RandomizedSearchCV`` would recompute the fingerprint for each set of hyperparameter
    values.

    Here, we instead perform a nested loop:

    1. Randomly select a combination of fingerprint hyperparameter values
    2. Compute fingerprint
    3. Optimize estimator hyperparameters

    This way, computed fingerprint representations are efficiently used for many sets of
    estimator hyperparameters. This is useful when tuning classifier or fingerprint and
    classifier. When only fingerprint is tuned, combination of ``Pipeline`` and
    ``GridSearchCV`` is enough. The difference is particularly significant for more
    computationally heavy fingerprints and large grids for estimators.

    Note that much of the behavior is controlled via passed ``estimator_cv`` object, e.g.
    the ``scoring`` metric used to select the best pipeline. In particular, the inner CV
    is evaluated for each one of ``n_iter`` random selections of the fingerprint
    hyperparameters, i.e. outer loop. This should be taken into consideration when
    selecting ``n_iter`` or hyperparameter grids. If ``RandomizedSearchCV`` is used, then
    the result is roughly equivalent to using randomized search on all hyperparameters,
    but faster. However, any other strategy can be used for the estimator,
    e.g. ``GridSearchCV``.

    Parameters
    ----------
    fingerprint : fingerprint object
        Instance of any fingerprint class. To maximize performance, consider setting
        ``n_jobs`` larger than 1, since parallelization is not performed here when going
        through fingerprint hyperparameter grid.

    fp_param_distributions : dict or list[dict]
        Dictionary with names of fingerprint hyperparameters as keys and lists of
        hyperparameter settings to try as values, or a list of such dictionaries, in which
        case the grids spanned by each dictionary in the list are explored. This enables
        searching over any sequence of hyperparameter settings.

    estimator_cv : object
        Inner cross-validation object for tuning estimator, e.g. ``RandomziedSearchCV``.
        Should be an instantiated object, not a class.

    greater_is_better : bool, default=True
        Whether higher values of scoring metric in ``estimator_cv`` are better or not.
        ``False`` should be used for error (loss) functions, typically used in regression.

    n_iter : int, default=10
        How many iterations of random search to perform.

    cache_best_fp_array : bool, default=False
        Whether to cache the array of values from the best fingerprint in ``best_fp_array_``
        parameter. Note that this can result in high memory usage.

    verbose : int, default=0
        Controls the verbosity: the higher, the more messages.

        - >0 : size of parameter grid, parameter candidate for each fold
        - >1 : the computation time and score for each candidate

    Attributes
    ----------
    cv_results_ : list[dict]
        List of dictionaries, where each one represents the set of hyperparameters
        (names and values) and ``"score"`` key with the cross-validated performance of the
        pipeline with those hyperparameters.

    best_fp_ : fingerprint object
        Fingerprint that was chosen by the search, i.e. fingerprint which gave the highest
        score (or smallest loss if specified) on the left out data. Use with
        `best_estimator_cv_` to obtain the best found pipeline.

    best_fp_params_ : dict
        Fingerprint hyperparameter values that gave the best results on the hold out data.

    best_fp_array_ : np.ndarray
        Fingerprint values for ``best_fp_``. If ``cache_best_fp_array`` is False, this will not
        be used and will be None instead.

    best_score_ : float
        Mean cross-validated score of the best fingerprint and estimator.

    best_estimator_cv_ : CV object
        Inner cross-validation object that gave the best results on the hold out data. Use
        with ``best_fp_`` to obtain the best found pipeline.

    See Also
    --------
    :class:`FingerprintEstimatorGridSearch` : Related fingerprint, but uses grid search
        for fingerprint hyperparameters.

    Examples
    --------
    >>> from skfp.datasets.moleculenet import load_bace
    >>> from skfp.fingerprints import ECFPFingerprint
    >>> from skfp.model_selection import FingerprintEstimatorRandomizedSearch
    >>> from sklearn.ensemble import RandomForestClassifier
    >>> from sklearn.model_selection import RandomizedSearchCV
    >>> smiles, labels = load_bace()
    >>> fp = ECFPFingerprint(n_jobs=-1)
    >>> fp_params = {"fp_size": list(range(512, 4097, 128))}
    >>> clf = RandomForestClassifier(n_jobs=-1)
    >>> clf_params = {"min_samples_split": list(range(2, 10))}
    >>> clf_cv = RandomizedSearchCV(clf, clf_params, n_iter=5, random_state=0)
    >>> fp_cv = FingerprintEstimatorRandomizedSearch(fp, fp_params, clf_cv, n_iter=5)
    >>> fp_cv = fp_cv.fit(smiles, labels)  # doctest: +SKIP
    >>> fp_cv.best_fp_params_  # doctest: +SKIP
    {'fp_size': 768}
    """

    _parameter_constraints: dict = {
        "fingerprint": [BaseFingerprintTransformer],
        "fp_param_distributions": [dict, list],
        "estimator_cv": [BaseSearchCV],
        "greater_is_better": ["boolean"],
        "n_iter": [Interval(Integral, 1, None, closed="left")],
        "cache_best_fp_array": ["boolean"],
        "verbose": ["verbose"],
        "random_state": ["random_state"],
    }

    def __init__(
        self,
        fingerprint: BaseFingerprintTransformer,
        fp_param_distributions: Union[dict, list[dict]],
        estimator_cv: BaseSearchCV,
        greater_is_better: bool = True,
        n_iter: int = 10,
        cache_best_fp_array: bool = False,
        verbose: int = 0,
        random_state: Optional[int] = 0,
    ):
        self.fingerprint = fingerprint
        self.fp_param_distributions = fp_param_distributions
        self.estimator_cv = estimator_cv
        self.greater_is_better = greater_is_better
        self.n_iter = n_iter
        self.cache_best_fp_array = cache_best_fp_array
        self.verbose = verbose
        self.random_state = random_state

    def _validate_params(self):
        super()._validate_params()
        if not self.fp_param_distributions:
            raise InvalidParameterError("fp_param_distributions cannot not be empty")

    @_fit_context(prefer_skip_nested_validation=False)
    def fit(self, X: Sequence[Union[str, Mol]], y=None, **params):
        self.cv_results_: list[dict] = []
        self.best_fp_: BaseFingerprintTransformer = None  # type: ignore
        self.best_fp_params_: dict = None  # type: ignore
        self.best_fp_array_: np.ndarray = None  # type: ignore
        self.best_score_ = None
        self.best_estimator_cv_: BaseSearchCV = None  # type: ignore

        param_sampler = ParameterSampler(
            self.fp_param_distributions, self.n_iter, random_state=self.random_state
        )
        if self.verbose:
            print(f"Fitting {self.n_iter} candidate hyperparameter sets.")

        for idx, fp_params in enumerate(param_sampler):
            if self.verbose:
                self._print_start_msg(
                    curr_idx=idx + 1,
                    curr_params=fp_params,
                )
            start_time = time()

            fp: BaseFingerprintTransformer = clone(self.fingerprint)
            fp.set_params(**fp_params)

            X_fp = fp.fit_transform(X, y, **params)

            curr_cv = clone(self.estimator_cv)
            curr_cv.fit(X_fp, y)
            curr_score = curr_cv.best_score_
            estimator_params = curr_cv.best_params_

            result = {**fp_params, **estimator_params, "score": curr_score}
            self.cv_results_.append(result)

            if (
                (self.best_score_ is None)
                or (self.greater_is_better and curr_score > self.best_score_)
                or (not self.greater_is_better and curr_score < self.best_score_)
            ):
                self.best_fp_ = fp
                self.best_fp_params_ = fp_params
                self.best_score_ = curr_score
                self.best_estimator_cv_ = curr_cv
                if self.cache_best_fp_array:
                    self.best_fp_array_ = X_fp

            end_time = time()
            if self.verbose > 1:
                self._print_end_msg(
                    curr_idx=idx + 1,
                    curr_params=fp_params,
                    start_time=start_time,
                    end_time=end_time,
                    curr_score=curr_score,
                )

        return self

    def predict(self, X: Sequence[Union[str, Mol]]) -> np.ndarray:
        X_fp = self.best_fp_.transform(X)
        return self.best_estimator_cv_.predict(X_fp)

    def predict_proba(self, X: Sequence[Union[str, Mol]]) -> np.ndarray:
        X_fp = self.best_fp_.transform(X)
        return self.best_estimator_cv_.predict_proba(X_fp)

    def transform(self, X: Sequence[Union[str, Mol]]) -> Union[np.ndarray, csr_array]:
        return self.best_fp_.transform(X)

    def _print_start_msg(self, curr_idx: int, curr_params: dict) -> None:
        progress_msg = f"{curr_idx}/{self.n_iter}"
        params_msg = ", ".join(
            f"{name}={curr_params[name]}" for name in sorted(curr_params)
        )
        start_msg = f"[{progress_msg}] START {params_msg}"
        print(f"{start_msg}{(80 - len(start_msg)) * '.'}")

    def _print_end_msg(
        self,
        curr_idx: int,
        curr_params: dict,
        start_time: float,
        end_time: float,
        curr_score: float,
    ) -> None:
        progress_msg = f"{curr_idx}/{self.n_iter}"
        params_msg = ", ".join(
            f"{name}={curr_params[name]}" for name in sorted(curr_params)
        )
        score_msg = f"score={curr_score:.3f}"
        total_time = joblib.logger.short_format_time(end_time - start_time)
        time_msg = f"total time={total_time}"
        end_msg = f"[{progress_msg}] END {params_msg}; {score_msg}; {time_msg}"
        print(f"{end_msg}{(80 - len(end_msg)) * '.'}")



from collections.abc import Sequence
from numbers import Integral
from typing import Optional, Union

import numpy as np
from rdkit.Chem import Mol
from scipy.sparse import csr_array
from sklearn.utils._param_validation import Interval, StrOptions

from skfp.bases import BaseFingerprintTransformer
from skfp.utils import ensure_mols


class PhysiochemicalPropertiesFingerprint(BaseFingerprintTransformer):
    """
    Physiochemical properties fingerprint.

    The implementation uses RDKit. Variants of this fingerprint [1]_ are
    binding property pairs (bp) and binding property torsions (bt), based on
    Atom Pairs and Topological Torsion fingerprint, respectively.

    The difference is in atom types (invariants) used, which here correspond
    to physiochemical features of an atom. They are therefore more general,
    but relate to functional role of atoms. Atom types are:
    - cation
    - anion
    - neutral hydrogen bond donor
    - neutral hydrogen bond acceptor
    - polar atom (both donor and acceptor, e.g. hydroxy oxygen)
    - hydrophobic
    - other

    Parameters
    ----------
    fp_size : int, default=2048
        Size of output vectors, i.e. number of bits for each fingerprint. Must be
        positive.

    variant: {"BP", "BT"}, default="BP"
        Which variant to use: binding property pairs `"BP"`, or binding property
        torsions `"BT"`.

    count : bool, default=False
        Whether to return binary (bit) features, or the count-based variant.

    sparse : bool, default=False
        Whether to return dense NumPy array, or sparse SciPy CSR array.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    Attributes
    ----------
    n_features_out : int
        Number of output features, size of fingerprints. Equal to ``fp_size``.

    requires_conformers : bool = False
        This fingerprint uses only 2D molecular graphs and does not require conformers.

    References
    ----------
    .. [1] `Simon K. Kearsley et al.
        "Chemical Similarity Using Physiochemical Property Descriptors"
        J. Chem. Inf. Comput. Sci. 1996, 36, 1, 118–127
        <https://pubs.acs.org/doi/10.1021/ci950274j>`_

    Examples
    --------
    >>> from skfp.fingerprints import PhysiochemicalPropertiesFingerprint
    >>> smiles = ["O", "CC", "[C-]#N", "CC=O"]
    >>> fp = PhysiochemicalPropertiesFingerprint()
    >>> fp
    PhysiochemicalPropertiesFingerprint()

    >>> fp.transform(smiles)
    array([[0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0]], dtype=uint8)
    """

    _parameter_constraints: dict = {
        **BaseFingerprintTransformer._parameter_constraints,
        "fp_size": [Interval(Integral, 1, None, closed="left")],
        "variant": [StrOptions({"BP", "BT"})],
    }

    def __init__(
        self,
        fp_size: int = 2048,
        variant: str = "BP",
        count: bool = False,
        sparse: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            n_features_out=fp_size,
            count=count,
            sparse=sparse,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )
        self.fp_size = fp_size
        self.variant = variant

    def _calculate_fingerprint(
        self, X: Sequence[Union[str, Mol]]
    ) -> Union[np.ndarray, csr_array]:
        from rdkit.Chem.AtomPairs.Sheridan import GetBPFingerprint, GetBTFingerprint

        X = ensure_mols(X)

        if self.variant == "BP":
            X = [GetBPFingerprint(mol) for mol in X]
        else:  # "BT" variant
            X = [GetBTFingerprint(mol) for mol in X]

        X = self._hash_fingerprint_bits(
            X, fp_size=self.fp_size, count=self.count, sparse=self.sparse
        )

        return X



from typing import Optional

from rdkit.Chem import FilterCatalog, Mol
from rdkit.Chem.rdfiltercatalog import FilterCatalogParams

from skfp.bases.base_filter import BaseFilter


class MLSMRFilter(BaseFilter):
    """
    MLSMR filter.

    NIH Molecular Libraries Small Molecule Repository (MLSMR) "Excluded Functionality
    Filters". MLSMR was a library of diverse small molecules (now discontinued).

    Filter origin and rules rationale are currently unknown, and there seems to be no
    literature reference. Rule definitions are available in the RDKit code [1]_.

    Parameters
    ----------
    allow_one_violation : bool, default=False
        Whether to allow violating one of the rules for a molecule. This makes the
        filter less restrictive.

    return_indicators : bool, default=False
        Whether to return a binary vector with indicators which molecules pass the
        filter, instead of list of molecules.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform_x_y` and
        :meth:`transform` are parallelized over the input molecules. ``None`` means 1
        unless in a :obj:`joblib.parallel_backend` context. ``-1`` means using all
        processors. See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when filtering molecules.

    References
    ----------
    .. [1] `RDKit MLSMR filter definitions
        <https://github.com/rdkit/rdkit/blob/e4f4644a89d6446ddebda0bf396fa4335324c41c/Code/GraphMol/FilterCatalog/chembl_mlsmr.in>`_

    Examples
    --------
    >>> from skfp.filters import MLSMRFilter
    >>> smiles = ["C", "O", "OP(=O)(O)O"]
    >>> filt = MLSMRFilter()
    >>> filt
    MLSMRFilter()

    >>> filtered_mols = filt.transform(smiles)
    >>> filtered_mols
    ['C', 'O']
    """

    def __init__(
        self,
        allow_one_violation: bool = False,
        return_indicators: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            allow_one_violation=allow_one_violation,
            return_indicators=return_indicators,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )
        self._filters = self._load_filters()

    def _load_filters(self) -> FilterCatalog:
        filter_rules = FilterCatalogParams.FilterCatalogs.CHEMBL_MLSMR
        params = FilterCatalog.FilterCatalogParams()
        params.AddCatalog(filter_rules)
        filters = FilterCatalog.FilterCatalog(params)
        return filters

    def _apply_mol_filter(self, mol: Mol) -> bool:
        errors = len(self._filters.GetMatches(mol))
        return not errors or (self.allow_one_violation and errors == 1)



from collections.abc import Sequence
from typing import Any

from rdkit.Chem import Mol, MolFromSmiles, MolToSmiles
from rdkit.Chem.PropertyMol import PropertyMol


def ensure_mols(X: Sequence[Any]) -> list[Mol]:
    if not all(isinstance(x, (Mol, PropertyMol, str)) for x in X):
        types = {type(x) for x in X}
        raise ValueError(
            f"Passed values must be one RDKit Mol objects or SMILES strings,"
            f"got types: {types}"
        )

    mols = [MolFromSmiles(x) if isinstance(x, str) else x for x in X]

    if any(x is None for x in mols):
        idx = mols.index(None)
        raise ValueError(f"Could not parse '{X[idx]}' at index {idx} as molecule")

    return mols


def ensure_smiles(X: Sequence[Any]) -> list[str]:
    if not all(isinstance(x, (Mol, PropertyMol, str)) for x in X):
        types = {type(x) for x in X}
        raise ValueError(f"Passed values must be SMILES strings, got types: {types}")

    X = [MolToSmiles(x) if isinstance(x, Mol) else x for x in X]
    return X


def check_strings(X: Sequence[Any]) -> None:
    for idx, x in enumerate(X):
        if not isinstance(x, str):
            raise ValueError(
                f"Passed values must be strings, got" f"type {type(x)} at index {idx}"
            )


def check_mols(X: Sequence[Any]) -> None:
    for idx, x in enumerate(X):
        if not isinstance(x, (Mol, PropertyMol)):
            raise ValueError(
                f"Passed values must be RDKit Mol objects, got type {type(x)} at index {idx}"
            )


def require_mols_with_conf_ids(X: Sequence[Any]) -> Sequence[Mol]:
    if not all(isinstance(x, (Mol, PropertyMol)) and x.HasProp("conf_id") for x in X):
        raise ValueError(
            "Passed data must be molecules (RDKit Mol objects) "
            "and each must have conf_id property set. "
            "You can use ConformerGenerator to add them."
        )
    return X



from typing import Optional

from rdkit.Chem import Mol
from rdkit.Chem.Crippen import MolLogP
from rdkit.Chem.Descriptors import MolWt
from rdkit.Chem.rdMolDescriptors import (
    CalcNumHBA,
    CalcNumHBD,
    CalcNumRotatableBonds,
    CalcTPSA,
)

from skfp.bases.base_filter import BaseFilter


class BeyondRo5Filter(BaseFilter):
    """
    Beyond Rule of Five (bRo5).

    Looser version of Lipinski's rule of 5, designed to cover novel orally bioavailable
    drugs that do not fulfill the original conditions [1]_. They are particularly
    suitable for "difficult" targets, allowing greater flexibility.

    Molecule can violate at most one of the rules (conditions):
    - molecular weight <= 1000 daltons
    - logP in range [-2, 10]
    - HBA <= 15
    - HBD <= 6
    - TPSA <= 250
    - number of rotatable bonds <= 20

    RDKit computes topological polar surface area (TPSA) in a specific way, differing
    slightly from published papers. See the documentation [2]_ for details.

    Parameters
    ----------
    allow_one_violation : bool, default=True
        Whether to allow violating one of the rules for a molecule. This makes the
        filter less restrictive.

    return_indicators : bool, default=False
        Whether to return a binary vector with indicators which molecules pass the
        filter, instead of list of molecules.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform_x_y` and
        :meth:`transform` are parallelized over the input molecules. ``None`` means 1
        unless in a :obj:`joblib.parallel_backend` context. ``-1`` means using all
        processors. See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when filtering molecules.

    References
    ----------
    .. [1] `Bradley C. Doak, Jie Zheng, Doreen Dobritzsch and Jan Kihlberg
        "How Beyond Rule of 5 Drugs and Clinical Candidates Bind to Their Targets"
        J. Med. Chem. 2016, 59, 6, 2312–2327
        <https://pubs.acs.org/doi/10.1021/acs.jmedchem.5b01286>`_

    .. [2] `RDKit Implementation of the TPSA Descriptor
        <https://www.rdkit.org/docs/RDKit_Book.html#implementation-of-the-tpsa-descriptor>`_

    Examples
    --------
    >>> from skfp.filters import BeyondRo5Filter, LipinskiFilter
    >>> smiles = ["[C-]#N", "CC=O", "O=C(O)C[C@H](O)C[C@H](O)CCn2c(c(c(c2c1ccc(F)cc1)c3ccccc3)C(=O)Nc4ccccc4)C(C)C"]
    >>> filt_bro5 = BeyondRo5Filter()
    >>> filt_bro5
    BeyondRo5Filter()

    >>> filt_bro5.transform(smiles)
    ['[C-]#N', 'CC=O', 'O=C(O)C[C@H](O)C[C@H](O)CCn2c(c(c(c2c1ccc(F)cc1)c3ccccc3)C(=O)Nc4ccccc4)C(C)C']

    >>> filt_ro5 = LipinskiFilter()
    >>> filt_ro5.transform(smiles)
    ['[C-]#N', 'CC=O']
    """

    def __init__(
        self,
        allow_one_violation: bool = False,
        return_indicators: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            allow_one_violation=allow_one_violation,
            return_indicators=return_indicators,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )

    def _apply_mol_filter(self, mol: Mol) -> bool:
        rules = [
            MolWt(mol) <= 1000,  # molecular weight
            -2 <= MolLogP(mol) <= 10,  # logP
            CalcNumHBA(mol) <= 15,  # HBA
            CalcNumHBD(mol) <= 6,  # HBD
            CalcTPSA(mol) <= 250,  # TPSA
            CalcNumRotatableBonds(mol) <= 20,
        ]
        passed_rules = sum(rules)

        if self.allow_one_violation:
            return passed_rules >= len(rules) - 1
        else:
            return passed_rules == len(rules)



from numbers import Real
from typing import Union

import numpy as np
from rdkit.ML.Scoring.Scoring import CalcBEDROC, CalcEnrichment, CalcRIE
from sklearn.utils import check_array, check_consistent_length
from sklearn.utils._param_validation import Interval, validate_params
from sklearn.utils.multiclass import type_of_target


@validate_params(
    {
        "y_true": ["array-like"],
        "y_pred": ["array-like"],
        "fraction": [Interval(Real, 0, 1, closed="neither")],
    },
    prefer_skip_nested_validation=True,
)
def enrichment_factor(
    y_true: Union[np.ndarray, list[int]],
    y_score: Union[np.ndarray, list[float]],
    fraction: float = 0.05,
) -> float:
    """
    Enrichment factor (EF).

    EF at fraction ``X`` is calculated as the number of actives found by the model,
    divided by the expected number of actives from a random ranking. See also [1]_
    for details.

    We have ``n`` actives, ``N`` test molecules, and percentage ``fraction`` of the
    top molecules (e.g. 0.05). Take ``fraction * N`` compounds with highest ``y_score``
    values, and mark the number of actives among them as ``a``. Random classifier
    would get on average ``fraction * n`` actives. Enrichment factor
    EF(X) is then defined as a ratio:

    .. math::

        EF(X) = \\frac{a}{X*n}

    Minimal value is 0. Maximal value depends on the fraction of actives in the
    dataset, and is equal to ``1/X`` if ``X >= n/N``, and ``N/n`` otherwise. Model
    as good as random guessing would get 1. Note that values depend on the ratio
    of actives in the dataset and ``fraction`` value.

    Parameters
    ----------
    y_true : array-like of shape (n_samples,)
        Ground truth (correct) target values.

    y_score : array-like of shape (n_samples,)
        Target scores, e.g. probability of the positive class, or similarities to
        active compounds.

    fraction : float, default=0.05
        Fraction of the dataset used for calculating the enrichment. Common values
        are 0.01 and 0.05. Note that this value affects the possible value range.

    Returns
    -------
    score : float
        Enrichment factor value.

    References
    ----------
    .. [1] `Riniker, S., Landrum, G.A.
        "Open-source platform to benchmark fingerprints for ligand-based virtual screening"
        J Cheminform 5, 26 (2013)
        <https://doi.org/10.1186/1758-2946-5-26>`_

    Examples
    --------
    >>> import numpy as np
    >>> from skfp.metrics import enrichment_factor
    >>> y_true = [0, 0, 1]
    >>> y_score = [0.1, 0.2, 0.7]
    >>> enrichment_factor(y_true, y_score)
    3.0
    """
    y_type = type_of_target(y_true, input_name="y_true")
    if y_type != "binary":
        raise ValueError(
            f"Enrichment factor is only defined for binary y_true, got {y_type}"
        )

    y_true = check_array(y_true, ensure_2d=False, dtype=None)
    y_score = check_array(y_score, ensure_2d=False)
    check_consistent_length(y_true, y_score)

    scores = sorted(zip(y_score, y_true), reverse=True)

    # RDKit returns a list, we have to extract the actual float
    # it can sometimes return an empty list, so we catch that and return 0.0 then, see:
    # https://github.com/rdkit/rdkit/issues/7981
    ef = CalcEnrichment(scores, col=1, fractions=[fraction])
    return 0.0 if not len(ef) else ef[0]


@validate_params(
    {
        "y_true": ["array-like"],
        "y_pred": ["array-like"],
        "alpha": [Interval(Real, 0, None, closed="neither")],
    },
    prefer_skip_nested_validation=True,
)
def rie_score(
    y_true: Union[np.ndarray, list[int]],
    y_score: Union[np.ndarray, list[float]],
    alpha: float = 20,
) -> float:
    """
    Robust Initial Enhancement (RIE).

    Exponentially weighted alternative to enrichment factor (EF), aimed to fix the
    problem of high variance with small number of actives. The exponential weight
    ``alpha`` is roughly equivalent to ``1 / fraction`` from EF. See [1]_ [2]_ [3]_
    for details.

    We have ``n`` actives, ``N`` test molecules, weight ``alpha``, and ``r_i`` is
    the rank of the i-th active from the test set. Then, RIE is defined as:

    .. math::

        RIE(\\alpha) = \\frac{N}{n}
            \\frac{\\sum_{i=1}^n e^{-\\alpha r_i / N}}
                  {\\frac{1 - e^{-\\alpha}}{e^{\\alpha/N} - 1}}

    Minimal and maximal value depend on ``n``, ``N`` and ``alpha``:

    .. math::

        RIE_{min}(\\alpha) = \\frac{N}{n} \\frac{1 - e^{\\alpha n/N}}{1 - e^{\\alpha}}

        RIE_{max}(\\alpha) = \\frac{N}{n} \\frac{1 - e^{-\\alpha n/N}}{1 - e^{-\\alpha}}

    Parameters
    ----------
    y_true : array-like of shape (n_samples,)
        Ground truth (correct) target values.

    y_score : array-like of shape (n_samples,)
        Target scores, e.g. probability of the positive class, or similarities to
        active compounds.

    alpha : float, default=20
        Exponential weight, roughly equivalent to ``1 / fraction`` from EF.

    References
    ----------
    .. [1] `Riniker, S., Landrum, G.A.
        "Open-source platform to benchmark fingerprints for ligand-based virtual screening"
        J Cheminform 5, 26 (2013).
        <https://doi.org/10.1186/1758-2946-5-26>`_

    .. [2] `Robert P. Sheridan et al.
        "Protocols for Bridging the Peptide to Nonpeptide Gap in Topological Similarity Searches"
        J. Chem. Inf. Comput. Sci. 2001, 41, 5, 1395–1406
        <https://doi.org/10.1021/ci0100144>`_

    .. [3] `Jean-François Truchon, Christopher I. Bayly
        "Evaluating Virtual Screening Methods: Good and Bad Metrics for the “Early Recognition” Problem"
        J. Chem. Inf. Model. 2007, 47, 2, 488–508
        <https://doi.org/10.1021/ci600426e>`_

    Returns
    -------
    score : float
        RIE value.

    Examples
    --------
    >>> import numpy as np
    >>> from skfp.metrics import rie_score
    >>> y_true = [0, 0, 1]
    >>> y_score = [0.1, 0.2, 0.7]
    >>> rie_score(y_true, y_score)  # doctest: +SKIP
    2.996182104771572
    """
    y_type = type_of_target(y_true, input_name="y_true")
    if y_type != "binary":
        raise ValueError(f"RIE is only defined for binary y_true, got {y_type}")

    y_true = check_array(y_true, ensure_2d=False, dtype=None)
    y_score = check_array(y_score, ensure_2d=False)
    check_consistent_length(y_true, y_score)

    scores = sorted(zip(y_score, y_true), reverse=True)
    return CalcRIE(scores, col=1, alpha=alpha)


@validate_params(
    {
        "y_true": ["array-like"],
        "y_pred": ["array-like"],
        "alpha": [Interval(Real, 0, None, closed="neither")],
    },
    prefer_skip_nested_validation=True,
)
def bedroc_score(
    y_true: Union[np.ndarray, list[int]],
    y_score: Union[np.ndarray, list[float]],
    alpha: float = 20,
) -> float:
    """
    Boltzmann-enhanced discrimination of ROC (BEDROC).

    A normalized alternative to Robust Initial Enhancement (RIE), scaled to have
    values in range ``[0, 1]``. ``alpha`` is directly passed to underlying RIE
    function, and is roughly equivalent to ``1 / fraction`` from enrichment factor
    (EF). See [1]_ [2]_ for details.

    Defined as:

    .. math::

        BEDROC(\\alpha) = \\frac{RIE(\\alpha) - RIE_{min}(\\alpha)}{RIE_{max}(\\alpha) - RIE_{min}(\\alpha)}

    Parameters
    ----------
    y_true : array-like of shape (n_samples,)
        Ground truth (correct) target values.

    y_score : array-like of shape (n_samples,)
        Target scores, e.g. probability of the positive class, or similarities to
        active compounds.

    alpha : float, default=20
        Exponential weight, roughly equivalent to ``1 / fraction`` from EF.

    References
    ----------
    .. [1] `Riniker, S., Landrum, G.A.
        "Open-source platform to benchmark fingerprints for ligand-based virtual screening"
        J Cheminform 5, 26 (2013).
        <https://doi.org/10.1186/1758-2946-5-26>`_

    .. [2] `Jean-François Truchon, Christopher I. Bayly
        "Evaluating Virtual Screening Methods: Good and Bad Metrics for the “Early Recognition” Problem"
        J. Chem. Inf. Model. 2007, 47, 2, 488–508
        <https://doi.org/10.1021/ci600426e>`_

    Returns
    -------
    score : float
        BEDROC value.

    Examples
    --------
    >>> import numpy as np
    >>> from skfp.metrics import bedroc_score
    >>> y_true = [0, 0, 1]
    >>> y_score = [0.1, 0.2, 0.7]
    >>> bedroc_score(y_true, y_score)  # doctest: +SKIP
    0.9999999999999999
    """
    y_type = type_of_target(y_true, input_name="y_true")
    if y_type != "binary":
        raise ValueError(f"BEDROC is only defined for binary y_true, got {y_type}")

    y_true = check_array(y_true, ensure_2d=False, dtype=None)
    y_score = check_array(y_score, ensure_2d=False)
    check_consistent_length(y_true, y_score)

    scores = sorted(zip(y_score, y_true), reverse=True)
    return CalcBEDROC(scores, col=1, alpha=alpha)



from typing import Optional

from rdkit.Chem import FilterCatalog, Mol
from rdkit.Chem.rdfiltercatalog import FilterCatalogParams
from sklearn.utils._param_validation import StrOptions

from skfp.bases.base_filter import BaseFilter


class PAINSFilter(BaseFilter):
    """
    Pan Assay Interference Compounds (PAINS) filter.

    Designed to filter out compounds that often give false positive results in
    high-throughput screens. Those are typically reactive molecules, which also
    contain functional groups resulting in high toxicity.

    There are 3 sets of PAINS filters, ordered by restrictiveness: A, B, C. PAINS A
    filters out the least molecules, while PAINS C is very aggressive. Rule definitions
    are available in the supplementary material of the original publication [1]_ (in
    Sybyl line notation), and in RDKit code [2]_ [3]_ [4]_ (as SMARTS patterns).

    Parameters
    ----------
    variant : {"A", "B", "C"}, default="A"
        PAINS filter set.

    allow_one_violation : bool, default=False
        Whether to allow violating one of the rules for a molecule. This makes the
        filter less restrictive.

    return_indicators : bool, default=False
        Whether to return a binary vector with indicators which molecules pass the
        filter, instead of list of molecules.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform_x_y` and
        :meth:`transform` are parallelized over the input molecules. ``None`` means 1
        unless in a :obj:`joblib.parallel_backend` context. ``-1`` means using all
        processors. See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when filtering molecules.

    References
    ----------
    .. [1] `Jonathan B. Baell and Georgina A. Holloway
        "New Substructure Filters for Removal of Pan Assay Interference Compounds
        (PAINS) from Screening Libraries and for Their Exclusion in Bioassays"
        J. Med. Chem. 2010, 53, 7, 2719–2740
        <https://doi.org/10.1021/jm901137j>`_

    .. [2] `RDKit PAINS A filter definitions
        <https://github.com/rdkit/rdkit/blob/e4f4644a89d6446ddebda0bf396fa4335324c41c/Code/GraphMol/FilterCatalog/pains_a.in>`_

    .. [3] `RDKit PAINS B filter definitions
        <https://github.com/rdkit/rdkit/blob/e4f4644a89d6446ddebda0bf396fa4335324c41c/Code/GraphMol/FilterCatalog/pains_b.in>`_

    .. [4] `RDKit PAINS C filter definitions
        <https://github.com/rdkit/rdkit/blob/e4f4644a89d6446ddebda0bf396fa4335324c41c/Code/GraphMol/FilterCatalog/pains_c.in>`_

    Examples
    --------
    >>> from skfp.filters import PAINSFilter
    >>> smiles = ["[C-]#N", "CC=O", "C1=CC(=O)C(=O)C=C1"]
    >>> filt = PAINSFilter()
    >>> filt
    PAINSFilter()

    >>> filtered_mols = filt.transform(smiles)
    >>> filtered_mols
    ['[C-]#N', 'CC=O']
    """

    _parameter_constraints = {
        **BaseFilter._parameter_constraints,
        "variant": [StrOptions({"A", "B", "C"})],
    }

    def __init__(
        self,
        variant: str = "A",
        allow_one_violation: bool = False,
        return_indicators: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            allow_one_violation=allow_one_violation,
            return_indicators=return_indicators,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )
        self.variant = variant
        self._filters = self._load_filters(variant)

    def _load_filters(self, variant: str) -> FilterCatalog:
        if variant == "A":
            filter_rules = FilterCatalogParams.FilterCatalogs.PAINS_A
        elif variant == "B":
            filter_rules = FilterCatalogParams.FilterCatalogs.PAINS_B
        elif variant == "C":
            filter_rules = FilterCatalogParams.FilterCatalogs.PAINS_C
        else:
            raise ValueError(f'PAINS variant must be "A", "B" or "C", got {variant}')

        params = FilterCatalog.FilterCatalogParams()
        params.AddCatalog(filter_rules)
        filters = FilterCatalog.FilterCatalog(params)
        return filters

    def _apply_mol_filter(self, mol: Mol) -> bool:
        errors = len(self._filters.GetMatches(mol))
        return not errors or (self.allow_one_violation and errors == 1)



from typing import Optional

from rdkit.Chem import Mol
from rdkit.Chem.Crippen import MolLogP
from rdkit.Chem.Descriptors import MolWt
from rdkit.Chem.rdchem import BondType
from rdkit.Chem.rdMolDescriptors import CalcNumHBA, CalcNumHBD, CalcNumRotatableBonds

from skfp.bases.base_filter import BaseFilter


class HaoFilter(BaseFilter):
    """
    Hao rule for pesticides.

    Designed to describe physicochemical properties of pesticides [1]_.
    Can be used in general pesticide design.

    Molecule must fulfill conditions:

        - molecular weight <= 435
        - logP <= 6
        - HBD <= 2
        - HBA <= 6
        - number of rotatable bonds <= 9
        - number of aromatic bonds <= 17


    Parameters
    ----------
    allow_one_violation : bool, default=False
        Whether to allow violating one of the rules for a molecule. This makes the
        filter less restrictive.

    return_indicators : bool, default=False
        Whether to return a binary vector with indicators which molecules pass the
        filter, instead of list of molecules.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform_x_y` and
        :meth:`transform` are parallelized over the input molecules. ``None`` means 1
        unless in a :obj:`joblib.parallel_backend` context. ``-1`` means using all
        processors. See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when filtering molecules.

    References
    -----------
    .. [1] `Hao, G., Dong, Q. and Yang, G.,
        "A Comparative Study on the Constitutive Properties of Marketed Pesticides."
        Mol. Inf., 30: 614-622.
        <https://doi.org/10.1002/minf.201100020>`_

    Examples
    ----------
    >>> from skfp.filters import HaoFilter
    >>> smiles = ["CN(C)c1ccc(C=Cc2cc[n+](C)c3ccccc23)cc1","c1cnc2c(c1)ccc1cccnc12",\
    "Cn1c(SSc2ccc(-c3cccnc3)n2C)ccc1-c1cccnc1"]
    >>> filt = HaoFilter()
    >>> filt
    HaoFilter()
    >>> filtered_mols = filt.transform(smiles)
    >>> filtered_mols
    ['CN(C)c1ccc(C=Cc2cc[n+](C)c3ccccc23)cc1', 'c1cnc2c(c1)ccc1cccnc12']
    """

    def __init__(
        self,
        allow_one_violation: bool = False,
        return_indicators: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            allow_one_violation=allow_one_violation,
            return_indicators=return_indicators,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )

    def _apply_mol_filter(self, mol: Mol) -> bool:
        aromatic_bond_count = sum(
            bond.GetBondType() == BondType.AROMATIC for bond in mol.GetBonds()
        )
        rules = [
            MolWt(mol) <= 435,
            MolLogP(mol) <= 6,
            CalcNumHBD(mol) <= 2,
            CalcNumHBA(mol) <= 6,
            CalcNumRotatableBonds(mol) <= 9,
            aromatic_bond_count <= 17,
        ]

        passed_rules = sum(rules)

        if self.allow_one_violation:
            return passed_rules >= len(rules) - 1
        else:
            return passed_rules == len(rules)



from typing import Optional

from rdkit.Chem import Mol
from rdkit.Chem.Crippen import MolLogP
from rdkit.Chem.Descriptors import MolWt
from rdkit.Chem.rdMolDescriptors import CalcNumHBA, CalcNumHBD, CalcNumRotatableBonds

from skfp.bases.base_filter import BaseFilter


class TiceHerbicidesFilter(BaseFilter):
    r"""
    Tice rule for herbicides.

    Rule established based on statistical analysis of herbicide molecules [1]_.
    Designed specifically for herbicides, not general pesticides or other agrochemicals.

    Molecule must fulfill conditions:

        - 150 <= molecular weight <= 500
        - logP <= 3.5
        - HBD <= 3
        - 2 <= HBA <= 12
        - number of rotatable bonds <= 11

    Parameters
    ----------
    allow_one_violation : bool, default=False
        Whether to allow violating one of the rules for a molecule. This makes the
        filter less restrictive.

    return_indicators : bool, default=False
        Whether to return a binary vector with indicators which molecules pass the
        filter, instead of list of molecules.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform_x_y` and
        :meth:`transform` are parallelized over the input molecules. ``None`` means 1
        unless in a :obj:`joblib.parallel_backend` context. ``-1`` means using all
        processors. See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when filtering molecules.

    References
    -----------
    .. [1] `Tice, C.M.
        "Selecting the right compounds for screening:
        does Lipinski's Rule of 5 for pharmaceuticals apply to agrochemicals?"
        Pest. Manag. Sci., 57: 3-16.
        <https://doi.org/10.1002/1526-4998(200101)57:1\<3::AID-PS269\>3.0.CO;2-6>`_

    Examples
    ----------
    >>> from skfp.filters import TiceHerbicidesFilter
    >>> smiles = ["OCCNc1nc2ccc(Cl)cc2[nH]1", "Nc1nnc(-c2ccco2)s1", "S=C1NCCS1"]
    >>> filt = TiceHerbicidesFilter()
    >>> filt
    TiceHerbicidesFilter()
    >>> filtered_mols = filt.transform(smiles)
    >>> filtered_mols
    ['OCCNc1nc2ccc(Cl)cc2[nH]1', 'Nc1nnc(-c2ccco2)s1']
    """

    def __init__(
        self,
        allow_one_violation: bool = False,
        return_indicators: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            allow_one_violation=allow_one_violation,
            return_indicators=return_indicators,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )

    def _apply_mol_filter(self, mol: Mol) -> bool:
        rules = [
            150 <= MolWt(mol) <= 500,
            MolLogP(mol) <= 3.5,
            CalcNumHBD(mol) <= 3,
            2 <= CalcNumHBA(mol) <= 12,
            CalcNumRotatableBonds(mol) <= 11,
        ]
        passed_rules = sum(rules)

        if self.allow_one_violation:
            return passed_rules >= len(rules) - 1
        else:
            return passed_rules == len(rules)



from typing import Callable, Union

import numpy as np
from sklearn.metrics import (
    accuracy_score,
    average_precision_score,
    balanced_accuracy_score,
    cohen_kappa_score,
    f1_score,
    matthews_corrcoef,
    mean_absolute_error,
    mean_squared_error,
    precision_score,
    recall_score,
    root_mean_squared_error,
)
from sklearn.utils._param_validation import validate_params

from skfp.metrics.auroc import auroc_score
from skfp.metrics.spearman import spearman_correlation


@validate_params(
    {
        "y_true": ["array-like"],
        "y_pred": ["array-like"],
    },
    prefer_skip_nested_validation=True,
)
def multioutput_accuracy_score(
    y_true: Union[np.ndarray, list],
    y_pred: Union[np.ndarray, list],
    *args,
    **kwargs,
) -> float:
    """
    Accuracy score for multioutput problems.

    Returns the average value over all tasks. Missing values in target labels are
    ignored. Also supports single-task evaluation.

    Any additional arguments are passed to the underlying ``accuracy_score`` function,
    see `scikit-learn documentation <sklearn>`_ for more information.

    .. _sklearn: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.accuracy_score.html

    Parameters
    ----------
    y_true : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Ground truth (correct) target values.

    y_pred : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Estimated target values.

    *args, **kwargs
        Any additional parameters for the underlying scikit-learn metric function.

    Returns
    -------
    score : float or int
        Average accuracy value over all tasks.

    Examples
    --------
    >>> import numpy as np
    >>> from skfp.metrics import multioutput_accuracy_score
    >>> y_true = [[0, 0], [1, 1]]
    >>> y_pred = [[0, 0], [0, 1]]
    >>> multioutput_accuracy_score(y_true, y_pred)
    0.75
    >>> y_true = [[0, np.nan], [1, np.nan], [np.nan, np.nan]]
    >>> y_pred = [[0, 0], [0, 0], [1, 0]]
    >>> multioutput_accuracy_score(y_true, y_pred)
    0.5
    """
    return _safe_multioutput_metric(accuracy_score, y_true, y_pred, *args, **kwargs)


@validate_params(
    {
        "y_true": ["array-like"],
        "y_score": ["array-like"],
    },
    prefer_skip_nested_validation=True,
)
def multioutput_auroc_score(
    y_true: Union[np.ndarray, list],
    y_score: Union[np.ndarray, list],
    *args,
    **kwargs,
) -> float:
    """
    Area Under Receiver Operating Characteristic curve (AUROC / ROC AUC) score for
    multioutput problems.

    Returns the average value over all tasks. Missing values in target labels are
    ignored. Columns with constant true value are ignored by default, but can also
    use default value - see ``auroc_score`` function. As such, it can be safely used
    e.g. in cross-validation. Also supports single-task evaluation.

    Any additional arguments are passed to the underlying ``auroc_score`` and ``roc_auc_score``
    functions, see `scikit-learn documentation <sklearn>`_ for more information.

    .. _sklearn: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.roc_auc_score.html

    Parameters
    ----------
    y_true : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Ground truth (correct) target values.

    y_score : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Target scores, i.e. probability of the class with the greater label for each
        output** of the classifier.

    *args, **kwargs
        Any additional parameters for the underlying scikit-learn metric function.

    Returns
    -------
    score : float
        Average AUROC value over all tasks.

    Examples
    --------
    >>> import numpy as np
    >>> from skfp.metrics import multioutput_auroc_score
    >>> y_true = np.array([[0, 0], [1, 1]])
    >>> y_score = np.array([[0.75, 0.0], [0.9, 0.0]])
    >>> multioutput_auroc_score(y_true, y_score)
    0.75
    >>> y_true = np.array([[0, 0], [1, np.nan], [np.nan, 1]])
    >>> y_score = np.array([[0.75, 0.0], [0.25, 0.0], [0.0, 0.25]])
    >>> multioutput_auroc_score(y_true, y_score)
    0.5
    """
    return _safe_multioutput_metric(auroc_score, y_true, y_score, *args, **kwargs)


@validate_params(
    {
        "y_true": ["array-like"],
        "y_score": ["array-like"],
    },
    prefer_skip_nested_validation=True,
)
def multioutput_auprc_score(
    y_true: Union[np.ndarray, list],
    y_score: Union[np.ndarray, list],
    *args,
    **kwargs,
) -> float:
    """
    Area Under Precision-Recall Curve (AUPRC / AUC PRC / average precision) score for
    multioutput problems.

    Returns the average value over all tasks. Missing values in target labels are
    ignored. Also supports single-task evaluation.

    Any additional arguments are passed to the underlying ``average_precision_score``
    function, see `scikit-learn documentation <sklearn>`_ for more information.

    .. _sklearn: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.average_precision_score.html

    Parameters
    ----------
    y_true : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Ground truth (correct) target values.

    y_score : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Target scores, i.e. probability of the class with the greater label for each
        output** of the classifier.

    *args, **kwargs
        Any additional parameters for the underlying scikit-learn metric function.

    Returns
    -------
    score : float
        Average AUPRC value over all tasks.

    Examples
    --------
    >>> import numpy as np
    >>> from skfp.metrics import multioutput_auprc_score
    >>> y_true = [[0, 0], [1, 1]]
    >>> y_score = [[0.75, 0.0], [0.9, 0.0]]
    >>> multioutput_auprc_score(y_true, y_score)
    0.75
    >>> y_true = [[0, 0], [1, np.nan], [np.nan, 1]]
    >>> y_score = [[0.75, 0.0], [0.25, 0.0], [0.0, 0.25]]
    >>> multioutput_auprc_score(y_true, y_score)
    0.75
    """
    return _safe_multioutput_metric(
        average_precision_score, y_true, y_score, *args, **kwargs
    )


@validate_params(
    {
        "y_true": ["array-like"],
        "y_pred": ["array-like"],
    },
    prefer_skip_nested_validation=True,
)
def multioutput_balanced_accuracy_score(
    y_true: Union[np.ndarray, list],
    y_pred: Union[np.ndarray, list],
    *args,
    **kwargs,
) -> float:
    """
    Balanced accuracy (average recall) score for multioutput problems.

    Returns the average value over all tasks. Missing values in target labels are
    ignored. Also supports single-task evaluation.

    Any additional arguments are passed to the underlying ``balanced_accuracy_score``
    function, see `scikit-learn documentation <sklearn>`_ for more information.

    .. _sklearn: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.balanced_accuracy_score.html

    Parameters
    ----------
    y_true : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Ground truth (correct) target values.

    y_pred : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Estimated target values.

    *args, **kwargs
        Any additional parameters for the underlying scikit-learn metric function.

    Returns
    -------
    score : float
        Average balanced accuracy value over all tasks.

    Examples
    --------
    >>> import numpy as np
    >>> from skfp.metrics import multioutput_balanced_accuracy_score
    >>> y_true = [[0, 0], [1, 1]]
    >>> y_pred = [[0, 0], [0, 1]]
    >>> multioutput_balanced_accuracy_score(y_true, y_pred)
    0.75
    >>> y_true = [[0, np.nan], [1, np.nan], [np.nan, np.nan]]
    >>> y_pred = [[0, 0], [0, 0], [1, 0]]
    >>> multioutput_balanced_accuracy_score(y_true, y_pred)
    0.5
    """
    return _safe_multioutput_metric(
        balanced_accuracy_score, y_true, y_pred, *args, **kwargs
    )


@validate_params(
    {
        "y_true": ["array-like"],
        "y_pred": ["array-like"],
    },
    prefer_skip_nested_validation=True,
)
def multioutput_cohen_kappa_score(
    y_true: Union[np.ndarray, list],
    y_pred: Union[np.ndarray, list],
    *args,
    **kwargs,
) -> float:
    """
    Cohen's kappa score for multioutput problems.

    Returns the average value over all tasks. Missing values in target labels are
    ignored. Also supports single-task evaluation.

    Any additional arguments are passed to the underlying ``cohen_kappa_score``
    function, see `scikit-learn documentation <sklearn>`_ for more information.

    .. _sklearn: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.cohen_kappa_score.html

    Parameters
    ----------
    y_true : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Ground truth (correct) target values.

    y_pred : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Estimated target values.

    *args, **kwargs
        Any additional parameters for the underlying scikit-learn metric function.

    Returns
    -------
    score : float
        Average Cohen's kappa value over all tasks.

    Examples
    --------
    >>> import numpy as np
    >>> from skfp.metrics import multioutput_cohen_kappa_score
    >>> y_true = [[0, 0], [1, 1]]
    >>> y_pred = [[0, 0], [0, 1]]
    >>> multioutput_cohen_kappa_score(y_true, y_pred)
    0.5
    >>> y_true = [[0, np.nan], [1, np.nan], [np.nan, np.nan]]
    >>> y_pred = [[0, 0], [0, 0], [1, 0]]
    >>> multioutput_cohen_kappa_score(y_true, y_pred)
    0.0
    """
    return _safe_multioutput_metric(cohen_kappa_score, y_true, y_pred, *args, **kwargs)


@validate_params(
    {
        "y_true": ["array-like"],
        "y_pred": ["array-like"],
    },
    prefer_skip_nested_validation=True,
)
def multioutput_f1_score(
    y_true: Union[np.ndarray, list],
    y_pred: Union[np.ndarray, list],
    *args,
    **kwargs,
) -> float:
    """
    F1 score for multioutput problems.

    Returns the average value over all tasks. Missing values in target labels are
    ignored. Columns with constant true value are also ignored, which differs from
    default scikit-learn behavior (it returns value 0 by default). Also supports
    single-task evaluation.

    Any additional arguments are passed to the underlying ``f1_score`` function,
    see `scikit-learn documentation <sklearn>`_ for more information.

    .. _sklearn: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.f1_score.html

    Parameters
    ----------
    y_true : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Ground truth (correct) target values.

    y_pred : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Estimated target values.

    *args, **kwargs
        Any additional parameters for the underlying scikit-learn metric function.

    Returns
    -------
    score : float
        Average F1 score value over all tasks.

    Examples
    --------
    >>> import numpy as np
    >>> from skfp.metrics import multioutput_f1_score
    >>> y_true = [[0, 0], [1, 1]]
    >>> y_pred = [[0, 0], [0, 1]]
    >>> multioutput_f1_score(y_true, y_pred)
    0.5
    >>> y_true = [[0, np.nan], [1, np.nan], [np.nan, np.nan]]
    >>> y_pred = [[0, 0], [0, 0], [1, 0]]
    >>> multioutput_f1_score(y_true, y_pred)
    0.0
    """
    return _safe_multioutput_metric(f1_score, y_true, y_pred, *args, **kwargs)


@validate_params(
    {
        "y_true": ["array-like"],
        "y_pred": ["array-like"],
    },
    prefer_skip_nested_validation=True,
)
def multioutput_matthews_corr_coef(
    y_true: Union[np.ndarray, list],
    y_pred: Union[np.ndarray, list],
    *args,
    **kwargs,
) -> float:
    """
    Matthews Correlation Coefficient (MCC) for multioutput problems.

    Returns the average value over all tasks. Missing values in target labels are
    ignored. Also supports single-task evaluation.

    Any additional arguments are passed to the underlying ``matthews_corrcoef``
    function, see `scikit-learn documentation <sklearn>`_ for more information.

    .. _sklearn: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.matthews_corrcoef.html

    Parameters
    ----------
    y_true : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Ground truth (correct) target values.

    y_pred : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Estimated target values.

    *args, **kwargs
        Any additional parameters for the underlying scikit-learn metric function.

    Returns
    -------
    score : float
        Average MCC value over all tasks.

    Examples
    --------
    >>> import numpy as np
    >>> from skfp.metrics import multioutput_matthews_corr_coef
    >>> y_true = [[0, 0], [1, 1]]
    >>> y_pred = [[0, 0], [0, 1]]
    >>> multioutput_matthews_corr_coef(y_true, y_pred)
    0.5
    >>> y_true = [[0, np.nan], [1, np.nan], [np.nan, np.nan]]
    >>> y_pred = [[0, 0], [0, 0], [1, 0]]
    >>> multioutput_matthews_corr_coef(y_true, y_pred)
    0.0
    """
    return _safe_multioutput_metric(matthews_corrcoef, y_true, y_pred, *args, **kwargs)


@validate_params(
    {
        "y_true": ["array-like"],
        "y_pred": ["array-like"],
    },
    prefer_skip_nested_validation=True,
)
def multioutput_mean_absolute_error(
    y_true: Union[np.ndarray, list],
    y_pred: Union[np.ndarray, list],
    *args,
    **kwargs,
) -> float:
    """
    Mean absolute error (MAE) for multioutput problems.

    Returns the average value over all tasks. Missing values in target labels are
    ignored. Also supports single-task evaluation.

    Any additional arguments are passed to the underlying ``mean_absolute_error``
    function, see `scikit-learn documentation <sklearn>`_ for more information.

    .. _sklearn: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_absolute_error.html

    Parameters
    ----------
    y_true : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Ground truth (correct) target values.

    y_pred : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Estimated target values.

    *args, **kwargs
        Any additional parameters for the underlying scikit-learn metric function.

    Returns
    -------
    score : float
        Average MAE value over all tasks.

    Examples
    --------
    >>> import numpy as np
    >>> from skfp.metrics import multioutput_mean_absolute_error
    >>> y_true = [[0.5, 1], [-1, 1], [7, -6]]
    >>> y_pred = [[0, 2], [-1, 2], [8, -5]]
    >>> multioutput_mean_absolute_error(y_true, y_pred)
    0.75
    >>> y_true = [[0.5, 1], [-1, 1], [np.nan, 10], [-10, np.nan]]
    >>> y_pred = [[0, 2], [-1, 2], [-3, 8], [-10, 5]]
    >>> multioutput_mean_absolute_error(y_true, y_pred)
    0.75
    """
    return _safe_multioutput_metric(
        mean_absolute_error, y_true, y_pred, *args, **kwargs
    )


@validate_params(
    {
        "y_true": ["array-like"],
        "y_pred": ["array-like"],
    },
    prefer_skip_nested_validation=True,
)
def multioutput_mean_squared_error(
    y_true: Union[np.ndarray, list],
    y_pred: Union[np.ndarray, list],
    *args,
    **kwargs,
) -> float:
    """
    Mean squared error (MSE) for multioutput problems.

    Returns the average value over all tasks. Missing values in target labels are
    ignored. Also supports single-task evaluation.

    Any additional arguments are passed to the underlying ``mean_squared_error``
    function, see `scikit-learn documentation <sklearn>`_ for more information.

    .. _sklearn: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_squared_error.html

    Parameters
    ----------
    y_true : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Ground truth (correct) target values.

    y_pred : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Estimated target values.

    *args, **kwargs
        Any additional parameters for the underlying scikit-learn metric function.

    Returns
    -------
    score : float
        Average MSE value over all tasks.

    Examples
    --------
    >>> import numpy as np
    >>> from skfp.metrics import multioutput_mean_squared_error
    >>> y_true = [[0.5, 1], [-1, 1], [7, -6]]
    >>> y_pred = [[0, 2], [-1, 2], [8, -5]]
    >>> multioutput_mean_squared_error(y_true, y_pred)
    0.708...
    >>> y_true = [[0.5, 1], [-1, 1], [np.nan, 10], [-10, np.nan]]
    >>> y_pred = [[0, 2], [-1, 2], [-3, 8], [-10, 5]]
    >>> multioutput_mean_squared_error(y_true, y_pred)
    1.041...
    """
    return _safe_multioutput_metric(mean_squared_error, y_true, y_pred, *args, **kwargs)


@validate_params(
    {
        "y_true": ["array-like"],
        "y_pred": ["array-like"],
    },
    prefer_skip_nested_validation=True,
)
def multioutput_precision_score(
    y_true: Union[np.ndarray, list],
    y_pred: Union[np.ndarray, list],
    *args,
    **kwargs,
) -> float:
    """
    Precision score for multioutput problems.

    Returns the average value over all tasks. Missing values in target labels are
    ignored. Warnings are not raised for columns with constant false prediction,
    and 0.0 is assumed. Also supports single-task evaluation.

    Any additional arguments are passed to the underlying ``precision_score``
    function, see `scikit-learn documentation <sklearn>`_ for more information.

    .. _sklearn: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.precision_score.html

    Parameters
    ----------
    y_true : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Ground truth (correct) target values.

    y_pred : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Estimated target values.

    *args, **kwargs
        Any additional parameters for the underlying scikit-learn metric function.

    Returns
    -------
    score : float
        Average precision value over all tasks.

    Examples
    --------
    >>> import numpy as np
    >>> from skfp.metrics import multioutput_precision_score
    >>> y_true = [[0, 0], [1, 1]]
    >>> y_pred = [[0, 0], [0, 1]]
    >>> multioutput_precision_score(y_true, y_pred)
    0.5
    >>> y_true = [[0, np.nan], [1, np.nan], [np.nan, np.nan]]
    >>> y_pred = [[0, 0], [0, 0], [1, 0]]
    >>> multioutput_precision_score(y_true, y_pred)
    0.0
    """
    return _safe_multioutput_metric(
        precision_score, y_true, y_pred, zero_division=0.0, *args, **kwargs
    )


@validate_params(
    {
        "y_true": ["array-like"],
        "y_pred": ["array-like"],
    },
    prefer_skip_nested_validation=True,
)
def multioutput_recall_score(
    y_true: Union[np.ndarray, list],
    y_pred: Union[np.ndarray, list],
    *args,
    **kwargs,
) -> float:
    """
    Recall score for multioutput problems.

    Returns the average value over all tasks. Missing values in target labels are
    ignored. Also supports single-task evaluation.

    Any additional arguments are passed to the underlying ``recall_score`` function,
    see `scikit-learn documentation <sklearn>`_ for more information.

    .. _sklearn: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.recall_score.html

    Parameters
    ----------
    y_true : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Ground truth (correct) target values.

    y_pred : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Estimated target values.

    *args, **kwargs
        Any additional parameters for the underlying scikit-learn metric function.

    Returns
    -------
    score : float
        Average recall value over all tasks.

    Examples
    --------
    >>> import numpy as np
    >>> from skfp.metrics import multioutput_recall_score
    >>> y_true = [[0, 0], [1, 1]]
    >>> y_pred = [[0, 0], [0, 1]]
    >>> multioutput_recall_score(y_true, y_pred)
    0.5
    >>> y_true = [[0, np.nan], [1, np.nan], [np.nan, np.nan]]
    >>> y_pred = [[0, 0], [0, 0], [1, 0]]
    >>> multioutput_recall_score(y_true, y_pred)
    0.0
    """
    return _safe_multioutput_metric(recall_score, y_true, y_pred, *args, **kwargs)


@validate_params(
    {
        "y_true": ["array-like"],
        "y_pred": ["array-like"],
    },
    prefer_skip_nested_validation=True,
)
def multioutput_root_mean_squared_error(
    y_true: Union[np.ndarray, list],
    y_pred: Union[np.ndarray, list],
    *args,
    **kwargs,
) -> float:
    """
    Root mean squared error (RMSE) for multioutput problems.

    Returns the average value over all tasks. Missing values in target labels are
    ignored. Also supports single-task evaluation.

    Any additional arguments are passed to the underlying ``root_mean_squared_error``
    function, see `scikit-learn documentation <sklearn>`_ for more information.

    .. _sklearn: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.root_mean_squared_error.html

    Parameters
    ----------
    y_true : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Ground truth (correct) target values.

    y_pred : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Estimated target values.

    *args, **kwargs
        Any additional parameters for the underlying scikit-learn metric function.

    Returns
    -------
    score : float
        Average RMSE value over all tasks.

    Examples
    --------
    >>> import numpy as np
    >>> from skfp.metrics import multioutput_root_mean_squared_error
    >>> y_true = [[0.5, 1], [-1, 1], [7, -6]]
    >>> y_pred = [[0, 2], [-1, 2], [8, -5]]
    >>> multioutput_root_mean_squared_error(y_true, y_pred)
    0.822...
    >>> y_true = [[0.5, 1], [-1, 1], [np.nan, 10], [-10, np.nan]]
    >>> y_pred = [[0, 2], [-1, 2], [-3, 8], [-10, 5]]
    >>> multioutput_root_mean_squared_error(y_true, y_pred)
    0.851...
    """
    return _safe_multioutput_metric(
        root_mean_squared_error, y_true, y_pred, *args, **kwargs
    )


@validate_params(
    {
        "y_true": ["array-like"],
        "y_pred": ["array-like"],
    },
    prefer_skip_nested_validation=True,
)
def multioutput_spearman_correlation(
    y_true: Union[np.ndarray, list],
    y_pred: Union[np.ndarray, list],
    *args,
    **kwargs,
) -> float:
    """
    Spearman correlation for multioutput problems.

    Returns the average value over all tasks. Missing values in target labels are
    ignored. Also supports single-task evaluation.

    Any additional arguments are passed to the underlying ``spearman_correlation``
    function, see :func:`spearman_correlation` for more information.

    Parameters
    ----------
    y_true : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Ground truth (correct) target values.

    y_pred : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Estimated target values.

    *args, **kwargs
        Any additional parameters for the underlying scikit-learn metric function.

    Returns
    -------
    score : float
        Average Spearman correlation value over all tasks.

    Examples
    --------
    >>> import numpy as np
    >>> from skfp.metrics import multioutput_spearman_correlation
    >>> y_true = np.array([[1, 1], [2, 2], [3, 3], [4, 4]])
    >>> y_pred = np.array([[1, 4], [2, 3], [3, 2], [4, 1]])
    >>> multioutput_spearman_correlation(y_true, y_pred)
    0.0
    >>> y_true = np.array([[1, 1], [np.nan, 2], [2, np.nan], [3, 3]])
    >>> y_pred = np.array([[1, 1], [0, 3], [3, 0], [4, 2]])
    >>> spearman_correlation(y_true, y_pred)
    0.75
    """
    return _safe_multioutput_metric(
        spearman_correlation, y_true, y_pred, *args, **kwargs
    )


def _safe_multioutput_metric(
    metric: Callable,
    y_true: Union[np.ndarray, list],
    y_pred: Union[np.ndarray, list],
    *args,
    **kwargs,
) -> float:
    if not isinstance(y_true, np.ndarray):
        y_true = np.array(y_true)
    if not isinstance(y_pred, np.ndarray):
        y_pred = np.array(y_pred)

    # make sure both arrays are 2D and have the same shape

    if y_true.shape != y_pred.shape:
        raise ValueError(
            f"Both true labels and predictions must have the same shape, got: "
            f"true labels {y_true.ndim}, predictions {y_pred.shape}"
        )

    if y_true.ndim == 1:
        y_true = y_true.reshape(-1, 1)
    elif y_true.ndim > 2:
        raise ValueError(
            f"True labels must have 1 or 2 dimensions, got shape {y_true.shape}"
        )

    if y_pred.ndim == 1:
        y_pred = y_pred.reshape(-1, 1)
    elif y_pred.ndim > 2:
        raise ValueError(
            f"Predictions must have 1 or 2 dimensions, got shape {y_true.shape}"
        )

    values = []
    for i in range(y_true.shape[1]):
        y_true_i = y_true[:, i]
        y_pred_i = y_pred[:, i]

        # in case of all-NaN column, skip it
        if np.all(np.isnan(y_true_i)):
            continue

        # remove NaN values
        non_nan_mask = ~np.isnan(y_true_i) & ~np.isnan(y_pred_i)
        y_true_i = y_true_i[non_nan_mask]
        y_pred_i = y_pred_i[non_nan_mask]

        col_value = metric(y_true_i, y_pred_i, *args, **kwargs)
        values.append(col_value)

    if not values:
        raise ValueError(
            "Could not compute metric value, y_true had only "
            "missing or constant values in all columns."
        )

    return np.mean(values)



from typing import Optional

from rdkit.Chem import FilterCatalog, Mol
from rdkit.Chem.rdfiltercatalog import FilterCatalogParams

from skfp.bases.base_filter import BaseFilter


class BMSFilter(BaseFilter):
    """
    BMS filter.

    Designed at BMS (Bristol-Myers Squibb) for filtering HTS decks in primary screening.
    Aims to remove molecules containing certin functional groups to filter out random
    noise, “promiscuous” compounds, and frequent hitters.

    Selected functional groups for filtering are divided in the paper into "exclusion
    lists" and "compound annotation". Which ones are "exclusion" filters is not clear
    from the paper, so this is a best-effort approximation from RDKit.

    Rule definitions are available in the supplementary material of the original
    publication [1]_ and in RDKit code [2]_. Note that

    Parameters
    ----------
    allow_one_violation : bool, default=False
        Whether to allow violating one of the rules for a molecule. This makes the
        filter less restrictive.

    return_indicators : bool, default=False
        Whether to return a binary vector with indicators which molecules pass the
        filter, instead of list of molecules.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform_x_y` and
        :meth:`transform` are parallelized over the input molecules. ``None`` means 1
        unless in a :obj:`joblib.parallel_backend` context. ``-1`` means using all
        processors. See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when filtering molecules.

    References
    ----------
    .. [1] `Bradley C. Pearce et al.
        "An Empirical Process for the Design of High-Throughput Screening Deck Filters"
        J. Chem. Inf. Model. 2006, 46, 3, 1060–1068
        <https://pubs.acs.org/doi/10.1021/ci050504m>`_

    .. [2] `RDKit BMS filter definitions
        <https://github.com/rdkit/rdkit/blob/e4f4644a89d6446ddebda0bf396fa4335324c41c/Code/GraphMol/FilterCatalog/chembl_bms.in>`_

    Examples
    --------
    >>> from skfp.filters import BMSFilter
    >>> smiles = ["C", "O", "c1(C=O)cc(OC)c(O)cc1"]
    >>> filt = BMSFilter()
    >>> filt
    BMSFilter()

    >>> filtered_mols = filt.transform(smiles)
    >>> filtered_mols
    ['C', 'O']
    """

    def __init__(
        self,
        allow_one_violation: bool = False,
        return_indicators: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            allow_one_violation=allow_one_violation,
            return_indicators=return_indicators,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )
        self._filters = self._load_filters()

    def _load_filters(self) -> FilterCatalog:
        filter_rules = FilterCatalogParams.FilterCatalogs.CHEMBL_BMS
        params = FilterCatalog.FilterCatalogParams()
        params.AddCatalog(filter_rules)
        filters = FilterCatalog.FilterCatalog(params)
        return filters

    def _apply_mol_filter(self, mol: Mol) -> bool:
        errors = len(self._filters.GetMatches(mol))
        return not errors or (self.allow_one_violation and errors == 1)



from collections.abc import Sequence
from numbers import Real
from typing import Optional, Union

import numpy as np
from rdkit.Chem import Mol
from scipy.sparse import csr_array
from sklearn.utils._param_validation import Interval

from skfp.bases import BaseFingerprintTransformer
from skfp.utils import require_mols_with_conf_ids


class GETAWAYFingerprint(BaseFingerprintTransformer):
    r"""
    GETAWAY (GEometry, Topology, and Atom-Weights AssemblY) fingerprint.

    The implementation uses RDKit. This is a descriptor-based fingerprint, where
    bits are features related to various autocorrelations and information measures
    defined on Molecular Influence Matrix (MIM).

    MIM matrix H is based on the centered atom coordinates (x,y,z) of a given
    conformer, and is therefore rotation invariant, and all features are independent
    of the conformer alignment. Diagonal elements of MIM matrix, called leverages,
    measure "influence" of each atom in determining the whole shape of the molecule.
    The influence matrix R, measuring strength of interatomic interactions, is then
    defined based on atom leverages and their spatial distances.

    GETAWAY descriptors consist of 273 features (see [3]_ [4]_ [5]_ [6]_ for precise
    definitions):

    - 7 related to general molecule shape, defined only on H and R matrices
    - 7 sets of autocorrelation descriptors, each defined on topological distances
      (shortest paths) from 0 to 8 (inclusive)

    Autocorrelation descriptors are unweighted, or weighted by: atomic mass, van der
    Waals volume, electronegativity, polarizability, ion polarity, and IState [1]_ [2]_.
    Those weights are relative to the carbon, e.g. molecular weight is: MW(atom_type) / MW(carbon).

    Typical correct values should be small, but it often results in NaN or infinity for
    some descriptors. Value clipping with ``clip_val`` parameter, feature selection, and/or
    imputation should be used.

    Parameters
    ----------
    clip_val : float or None, default=2147483647
        Value to clip results at, both positive and negative ones.The default value is
        the maximal value of 32-bit integer, but should often be set lower, depending
        on the application. ``None`` means that no clipping is applied.

    sparse : bool, default=False
        Whether to return dense NumPy array, or sparse SciPy CSR array.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    Attributes
    ----------
    n_features_out : int = 273
        Number of output features, size of fingerprints.

    requires_conformers : bool = True
        Value is always True, as this fingerprint is 3D based. It always requires
        molecules with conformers as inputs, with ``conf_id`` integer property set.

    References
    ----------
    .. [1] RDKit IState discussion
        https://github.com/rdkit/rdkit/discussions/6122

    .. [2] RDKit IState implementation
        https://github.com/rdkit/rdkit/blob/df2b6c8a8c775748c1dcac83af0f92af434bab81/Code/GraphMol/Descriptors/MolData3Ddescriptors.cpp#L127

    .. [3] `Rajarshi Guha
        "GETAWAY"
        <http://www.rguha.net/writing/notes/desc/node5.html>`_

    .. [4] `Roberto Todeschini and Viviana Consonni
        "Molecular Descriptors for Chemoinformatics"
        <https://onlinelibrary.wiley.com/doi/book/10.1002/9783527628766>`_

    .. [5] `Guillaume Godin
        "3D	descriptors in RDKit"
        UGM 2017
        <https://github.com/rdkit/UGM_2017/blob/master/Presentations/Godin_3D_Descriptors.pdf>`_

    .. [6] `Viviana Consonni, Roberto Todeschini, and Manuela Pavan
        "Structure/Response Correlations and Similarity/Diversity Analysis by GETAWAY Descriptors.
        1. Theory of the Novel 3D Molecular Descriptors"
        J. Chem. Inf. Comput. Sci. 2002, 42, 3, 682–692
        <https://pubs.acs.org/doi/abs/10.1021/ci015504a>`_

    Examples
    --------
    >>> from skfp.fingerprints import GETAWAYFingerprint
    >>> from skfp.preprocessing import MolFromSmilesTransformer, ConformerGenerator
    >>> smiles = ["O", "CC", "[C-]#N", "CC=O"]
    >>> fp = GETAWAYFingerprint()
    >>> fp
    GETAWAYFingerprint()

    >>> mol_from_smiles = MolFromSmilesTransformer()
    >>> mols = mol_from_smiles.transform(smiles)
    >>> conf_gen = ConformerGenerator()
    >>> mols = conf_gen.transform(mols)
    >>> fp.transform(mols)  # doctest: +SKIP
    array([[ 0.   ,    nan,  1.585, ...,  1.   , -0.   ,  1.   ],
           [ 0.   ,  0.   ,  2.763, ...,  1.   ,  0.   ,  1.   ],
           [ 0.   ,  0.   ,  1.   , ...,  1.   ,  0.   , 13.076],
           [ 4.755,  1.   ,  2.502, ..., -1.   , -0.   ,  2.467]])
    """

    _parameter_constraints: dict = {
        **BaseFingerprintTransformer._parameter_constraints,
        "clip_val": [None, Interval(Real, 0, None, closed="left")],
    }

    def __init__(
        self,
        clip_val: float = 2147483647,  # max int32 value
        sparse: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            n_features_out=273,
            requires_conformers=True,
            sparse=sparse,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )
        self.clip_val = clip_val

    def _calculate_fingerprint(self, X: Sequence[Mol]) -> Union[np.ndarray, csr_array]:
        from rdkit.Chem.rdMolDescriptors import CalcGETAWAY

        X = require_mols_with_conf_ids(X)
        X = [CalcGETAWAY(mol, confId=mol.GetIntProp("conf_id")) for mol in X]
        if self.clip_val is not None:
            X = np.clip(X, -self.clip_val, self.clip_val)

        return csr_array(X) if self.sparse else np.array(X)



"""Classes for computing molecular fingerprints."""

from .atom_pair import AtomPairFingerprint
from .autocorr import AutocorrFingerprint
from .avalon import AvalonFingerprint
from .e3fp_fp import E3FPFingerprint
from .ecfp import ECFPFingerprint
from .electroshape import ElectroShapeFingerprint
from .erg import ERGFingerprint
from .estate import EStateFingerprint
from .functional_groups import FunctionalGroupsFingerprint
from .getaway import GETAWAYFingerprint
from .ghose_crippen import GhoseCrippenFingerprint
from .klekota_roth import KlekotaRothFingerprint
from .laggner import LaggnerFingerprint
from .layered import LayeredFingerprint
from .lingo import LingoFingerprint
from .maccs import MACCSFingerprint
from .map import MAPFingerprint
from .mhfp import MHFPFingerprint
from .mordred_fp import MordredFingerprint
from .morse import MORSEFingerprint
from .mqns import MQNsFingerprint
from .pattern import PatternFingerprint
from .pharmacophore import PharmacophoreFingerprint
from .physiochemical_properties import PhysiochemicalPropertiesFingerprint
from .pubchem import PubChemFingerprint
from .rdf import RDFFingerprint
from .rdkit_2d_desc import RDKit2DDescriptorsFingerprint
from .rdkit_fp import RDKitFingerprint
from .secfp import SECFPFingerprint
from .topological_torsion import TopologicalTorsionFingerprint
from .usr import USRFingerprint
from .usrcat import USRCATFingerprint
from .vsa import VSAFingerprint
from .whim import WHIMFingerprint



from typing import Optional

from rdkit.Chem import Mol
from rdkit.Chem.Crippen import MolLogP
from rdkit.Chem.Descriptors import MolWt
from rdkit.Chem.rdMolDescriptors import (
    CalcNumHBA,
    CalcNumHBD,
    CalcNumRotatableBonds,
    CalcTPSA,
)

from skfp.bases.base_filter import BaseFilter


class RuleOfThreeFilter(BaseFilter):
    """
    Rule of three (Ro3).

    Rule optimised to search for fragment-based lead-like compounds with desired properties.
    It was described in [1]_.

    Molecule must fulfill conditions:

    - molecular weight <= 300 daltons
    - HBA <= 3
    - HBD <= 3
    - logP <= 3

    Additionally, an extended version of this rule has been proposed, which adds two conditions:

    - TPSA <= 60
    - number of rotatable bonds <= 3

    Parameters
    ----------
    extended : bool, default=False
        Whether to use an extended version of this rule, additionally including TPSA and
        rotatable bonds conditions.

    allow_one_violation : bool, default=False
        Whether to allow violating one of the rules for a molecule. This makes the
        filter less restrictive.

    return_indicators : bool, default=False
        Whether to return a binary vector with indicators which molecules pass the
        filter, instead of list of molecules.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform_x_y` and
        :meth:`transform` are parallelized over the input molecules. ``None`` means 1
        unless in a :obj:`joblib.parallel_backend` context. ``-1`` means using all
        processors. See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when filtering molecules.

    References
    -----------
    .. [1] `Congreve, M., Carr, R., Murray, C., & Jhoti, H.
        "A 'Rule of Three' for fragment-based lead discovery?"
        Drug Discovery Today, 8(19), 876–877.
        <https://doi.org/10.1016/S1359-6446(03)02831-9>`_

    Examples
    ----------
    >>> from skfp.filters import RuleOfThreeFilter
    >>> smiles = ['C=CCNC(=S)NCc1ccccc1OC', 'C=CCOc1ccc(Br)cc1/C=N/O', 'C=CCNc1ncnc2ccccc12']
    >>> filt = RuleOfThreeFilter()
    >>> filt
    RuleOfThreeFilter()
    >>> filtered_mols = filt.transform(smiles)
    >>> filtered_mols
    ['C=CCNC(=S)NCc1ccccc1OC', 'C=CCOc1ccc(Br)cc1/C=N/O', 'C=CCNc1ncnc2ccccc12']
    >>> filt = RuleOfThreeFilter(extended=True)
    >>> filtered_mols = filt.transform(smiles)
    >>> filtered_mols
    ['C=CCNc1ncnc2ccccc12']
    """

    def __init__(
        self,
        extended: bool = False,
        allow_one_violation: bool = False,
        return_indicators: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            allow_one_violation=allow_one_violation,
            return_indicators=return_indicators,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )
        self.extended = extended

    def _apply_mol_filter(self, mol: Mol) -> bool:
        rules = [
            MolWt(mol) <= 300,
            CalcNumHBA(mol) <= 3,
            CalcNumHBD(mol) <= 3,
            MolLogP(mol) <= 3,
        ]
        if self.extended:
            rules += [CalcNumRotatableBonds(mol) <= 3, CalcTPSA(mol) <= 60]
        passed_rules = sum(rules)

        if self.allow_one_violation:
            return passed_rules >= len(rules) - 1
        else:
            return passed_rules == len(rules)



from typing import Union

import numpy as np
from sklearn.utils._param_validation import validate_params


@validate_params(
    {"predictions": ["array-like"]},
    prefer_skip_nested_validation=True,
)
def extract_pos_proba(predictions: Union[np.ndarray, list[np.ndarray]]) -> np.ndarray:
    """
    Extract positive class probabilities (``y-score``).

    Probabilitic metrics like AUROC or AUPRC use predicted probabilities. This
    function extracts them from ``.predict_proba()`` results.

    Returns `(n_samples,)` shape for single-task, and `(n_samples, n_tasks)` for
    multioutput problems.

    Parameters
    ----------
    predictions : list of NumPy arrays
        Raw predictions of shape ``(n_samples,2)`` (single-task) or
        ``(n_tasks, n_samples, 2)`` (multioutput), with predicted negative
        and positive class probability in the last dimension.

    Returns
    -------
    y_score : NumPy array of shape (n_samples,) or (n_samples, n_tasks
        Predicted positive class probabilities for each task.

    Examples
    --------
    >>> import numpy as np
    >>> from skfp.metrics import extract_pos_proba
    >>> y_pred = np.array([[0.6, 0.1], [0.2, 0.3]])
    >>> extract_pos_proba(y_pred)
    array([0.1, 0.3])
    >>> y_pred = [np.array([[0.6, 0.1], [0.2, 0.3]]), np.array([[0.0, 0.9], [0.7, 0.8]])]
    >>> extract_pos_proba(y_pred)
    array([[0.1, 0.9],
           [0.3, 0.8]])
    """
    # (n_tasks, n_samples, 2) -> (n_tasks, n_samples) -> (n_samples, n_tasks)
    predictions = np.array(predictions)
    if predictions.ndim == 2:
        return predictions[:, 1]
    elif predictions.ndim == 3:
        return np.transpose(predictions[:, :, 1])
    else:
        raise ValueError(
            f"Predictions must have 2 or 3 dimensions, got {predictions.ndim}"
        )



import hashlib
import re
from collections import defaultdict
from collections.abc import Sequence
from numbers import Integral
from typing import Optional, Union

import numpy as np
from rdkit.Chem import Mol
from scipy.sparse import csr_array
from sklearn.utils._param_validation import Interval

from skfp.bases import BaseFingerprintTransformer
from skfp.utils import ensure_smiles


class LingoFingerprint(BaseFingerprintTransformer):
    """
    Lingo fingerprint.

    The Lingo fingerprint [1]_ is a hashed fingerprint that checks the occurrences of
    substrings of a given length in a SMILES string. These strings are overlapping.
    Original paper uses raw dictionaries os output, but here constant-length hashed
    fingerprints are returned. SHA-1 hash function is used here.

    You can use method :meth:`smiles_to_dicts` to convert SMILES strings to
    dictionaries of substring, like original Lingo fingerprint.

    Parameters
    ----------
    fp_size : int, default=1024
        Size of output vectors, i.e. number of bits for each fingerprint. Must be
        positive.

    substring_length : int, default=4
        Length of the substrings to count in the SMILES strings. Must be positive.

    count : bool, default=False
        Whether to return binary (bit) features, or their counts.

    sparse : bool, default=False
        Whether to return dense NumPy array, or sparse SciPy CSR array.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    Attributes
    ----------
    n_features_out : int
        Number of output features, size of fingerprints. Equal to ``fp_size``.

    References
    ----------
    .. [1] `David Vidal, Michael Thormann, and Miquel Pons
        "LINGO, an Efficient Holographic Text Based Method To Calculate
        Biophysical Properties and Intermolecular Similarities"
        J. Chem. Inf. Model. 2005, 45, 2, 386-393
        <https://pubs.acs.org/doi/abs/10.1021/ci0496797>`_

    Examples
    --------
    >>> from skfp.fingerprints import LingoFingerprint
    >>> smiles = ["[C-]#N", "CC(=O)NC1=CC=C(C=C1)O"]
    >>> fp = LingoFingerprint()
    >>> fp
    LingoFingerprint()
    >>> fp.transform(smiles)
    array([[0, 0, 0, ..., 0, 0, 0],
           [0, 1, 0, ..., 0, 0, 0]], dtype=uint8)
    """

    _parameter_constraints: dict = {
        **BaseFingerprintTransformer._parameter_constraints,
        "fp_size": [Interval(Integral, 1, None, closed="left")],
        "substring_length": [Interval(Integral, 1, None, closed="left")],
        "count": ["boolean"],
    }

    def __init__(
        self,
        fp_size: int = 1024,
        substring_length: int = 4,
        count: bool = False,
        sparse: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            n_features_out=fp_size,
            count=count,
            sparse=sparse,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )
        self.fp_size = fp_size
        self.substring_length = substring_length

    def transform(
        self, X: Sequence[Union[str, Mol]], copy: bool = False
    ) -> Union[np.ndarray, csr_array]:
        """
        Compute Lingo fingerprints.

        Parameters
        ----------
        X : {sequence, array-like} of shape (n_samples,)
            Sequence containing SMILES strings or RDKit Mol objects.

        copy : bool, default=False
            Whether to copy X or modify it in place.

        Returns
        -------
        X: {ndarray, sparse matrix} of shape (n_samples, self.fp_size)
            2D array containing Lingo fingerprints.
        """
        return super().transform(X, copy)

    def smiles_to_dicts(self, X: Sequence[Union[str, Mol]]) -> list[dict[str, int]]:
        """
        Convert SMILES strings to dictionaries of substring counts.

        Parameters
        ----------
        X: {sequence, array-like} of shape (n_samples,)
            Sequence containing SMILES strings or RDKit Mol objects.

        Returns
        -------
        result: list[dict[str, int]]
            List of dictionaries containing substring counts.
        """
        X = ensure_smiles(X)

        # based on the original paper, we reduce the number of possible substrings
        # to improve statistical sampling in the QSPR models
        X = [re.sub(r"[123456789]", "0", smi) for smi in X]
        X = [re.sub(r"Cl", "L", smi) for smi in X]
        X = [re.sub(r"Br", "R", smi) for smi in X]

        result = []
        for smi in X:
            result_dict: defaultdict[str, int] = defaultdict(int)
            for i in range(len(smi) - self.substring_length + 1):
                result_dict[smi[i : i + self.substring_length]] += 1
            result.append(dict(result_dict))

        return result

    def _calculate_fingerprint(
        self, X: Sequence[Union[str, Mol]]
    ) -> Union[np.ndarray, csr_array]:
        X = self.smiles_to_dicts(X)
        X = self._dicts_to_array(X)
        return csr_array(X) if self.sparse else X

    def _dicts_to_array(self, X: list[dict[str, int]]) -> np.ndarray:
        dtype = np.uint32 if self.count else np.uint8
        result = np.zeros((len(X), self.fp_size), dtype=dtype)

        for i, dictionary in enumerate(X):
            for key, value in dictionary.items():
                string_bytes = key.encode("utf-8")
                hash_bytes = hashlib.sha1(string_bytes, usedforsecurity=False).digest()
                hash_index = int.from_bytes(hash_bytes, byteorder="big") % self.fp_size
                if self.count:
                    result[i, hash_index] += value
                else:
                    result[i, hash_index] = 1

        return result



from collections.abc import Sequence
from copy import deepcopy
from functools import partial
from numbers import Integral
from typing import Optional

import numpy as np
from rdkit.Chem import AddHs, Mol, MolToSmiles, RemoveHs, SanitizeMol
from rdkit.Chem.PropertyMol import PropertyMol
from rdkit.Chem.rdDistGeom import (
    EmbedFailureCauses,
    EmbedMolecule,
    EmbedMultipleConfs,
    ETKDGv3,
)
from rdkit.Chem.rdForceFieldHelpers import (
    MMFFGetMoleculeForceField,
    MMFFGetMoleculeProperties,
    MMFFSanitizeMolecule,
    UFFGetMoleculeForceField,
)
from rdkit.ForceField import ForceField
from sklearn.utils._param_validation import Interval, InvalidParameterError, StrOptions

from skfp.bases import BasePreprocessor


class ConformerGenerator(BasePreprocessor):
    """
    Generate molecule conformer.

    The implementation uses RDKit and distance geometry (DG) approach [1]_, with
    optimized ETKDGv3 improvements for small rings, macrocycles and experimental
    torsional angle preferences [2]_. Generated conformations are optionally optimized
    with a force field approach.

    Resulting conformation is saved in ``conf_id`` integer property of a molecule, and
    can be retrieved with ``GetIntProp("conf_id")`` method.

    If multiple conformations are generated, one per molecule is returned. By default,
    the most stable conformer (with the lowest energy) is selected.

    Note that conformer generation can fail, either due to not enough iterations, or
    it can be just impossible for a given molecule [3]_. This by default results in an
    error, but can be controlled with ``error_on_gen_fail`` parameter. For multiple
    conformers, error is thrown only when no conformations can be generated, not if
    any one fails.

    If ``error_on_gen_fail`` is False and no conformers could be generated, the number
    of returned molecules will be smaller than input length. For supervised learning,
    use ``transform_x_y()`` instead of ``transform`` method to properly return labels
    for those molecules.

    Parameters
    ----------
    num_conformers : int, default=1
        Number of conformers to initially generate for each molecule. Must be positive.
        If larger than 1, one conformation will be selected, as specified with
        ``multiple_confs_select``.

    max_gen_attempts : int, default=10000
        Number of attempts to generate a conformer. Must be positive, and should be
        sufficiently large, typically at least a few thousands.

    optimize_force_field : {"UFF", "MMFF94", "MMFF94s", None}, default=None
        Force field optimization algorithm used on generated conformers. It is also
        used for calculation of conformer energy when selecting one of multiple
        conformations with ``multiple_confs_select="min_energy"``.

    multiple_confs_select : {"min_energy", "first"}, default="min_energy"
        How to select final conformer for each molecule when multiple conformers are
        generated. "first" selects first conformer generated by RDKit.

    errors : {"raise", "ignore", "filter"}, default="raise"
        How to handle errors during conformer generation, if after all attempts and
        approaches a conformer could not be generated. ``"raise"`` immediately raises
        any errors, with failure reason. ``"ignore"`` suppresses errors and returns
        molecules without conformers and ``conf_id`` property set to `-1`. ``"filter"``
        suppresses errors and removes such conformers, resulting in potentially less
        output molecules than inputs. The latter two options can potentially cause
        problems downstream, and should be used with caution.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform_x_y` and
        :meth:`transform` are parallelized over the input molecules. ``None`` means 1
        unless in a :obj:`joblib.parallel_backend` context. ``-1`` means using all
        processors. See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when generating conformers.

    random_state : int or None, default=0
        Controls the randomness of conformer generation. Note that in constrast to
        most classes, here it cannot be a ``RandomState`` instance, only an integer.

    References
    ----------
    .. [1] `Jean-Paul Ebejer
        "Conformer Generation using RDKit"
        1st RDKit User General Meeting, London, 2012
        <https://www.rdkit.org/UGM/2012/Ebejer_20110926_RDKit_1stUGM.pdf>`_

    .. [2] `Shuzhe Wang, Jagna Witek, Gregory A. Landrum, and Sereina Riniker
        "Improving Conformer Generation for Small Rings and Macrocycles Based on
        Distance Geometry and Experimental Torsional-Angle Preferences"
        J. Chem. Inf. Model. 2020, 60, 4, 2044–2058
        <https://pubs.acs.org/doi/10.1021/acs.jcim.0c00025>`_

    .. [3] `Gregory A. Landrum
        "Understanding conformer generation failures"
        RDKit blog, 2023
        <https://greglandrum.github.io/rdkit-blog/posts/2023-05-17-understanding-confgen-errors.html>`_

    Examples
    --------
    >>> from skfp.preprocessing import MolFromSmilesTransformer, ConformerGenerator
    >>> smiles = ["O", "CC", "[C-]#N", "CC=O"]
    >>> conf_gen = ConformerGenerator()
    >>> conf_gen
    ConformerGenerator()

    >>> mol_from_smiles = MolFromSmilesTransformer()
    >>> mols = mol_from_smiles.transform(smiles)
    >>> conf_gen.transform(mols) # doctest: +ELLIPSIS
    [<rdkit.Chem.PropertyMol.PropertyMol...>, ..., <rdkit.Chem.PropertyMol.PropertyMol...>]
    """

    _parameter_constraints: dict = {
        "num_conformers": [Interval(Integral, 1, None, closed="left")],
        "max_gen_attempts": [Interval(Integral, 1, None, closed="left")],
        "optimize_force_field": [StrOptions({"UFF", "MMFF94", "MMFF94s"}), None],
        "multiple_confs_select": [StrOptions({"min_energy", "first"})],
        "errors": [StrOptions({"raise", "ignore", "filter"})],
        "n_jobs": [Integral, None],
        "batch_size": [Integral, None],
        "verbose": ["verbose"],
        "random_state": [Interval(Integral, left=-1, right=None, closed="left")],
    }

    def __init__(
        self,
        num_conformers: int = 1,
        max_gen_attempts: int = 1000,
        optimize_force_field: Optional[str] = None,
        multiple_confs_select: Optional[str] = "min_energy",
        errors: str = "raise",
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
        random_state: Optional[int] = 0,
    ):
        super().__init__(
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )
        self.num_conformers = num_conformers
        self.max_gen_attempts = max_gen_attempts
        self.optimize_force_field = optimize_force_field
        self.multiple_confs_select = multiple_confs_select
        self.errors = errors
        self.random_state = random_state

    def _validate_params(self) -> None:
        super()._validate_params()
        if (
            self.num_conformers > 1
            and self.multiple_confs_select == "min_energy"
            and self.optimize_force_field is None
        ):
            raise InvalidParameterError(
                "For selecting one of multiple conformers with lowest energy "
                '(num_conformers > 1 and multiple_confs_select == "min_energy"), '
                "force field optimization algorithm must be selected with "
                "optimize_force_field parameter, got None."
            )

    def transform_x_y(
        self, X: Sequence[Mol], y: np.ndarray, copy: bool = False
    ) -> tuple[list[PropertyMol], np.ndarray]:
        """
        Generate conformers for molecules.

        If ``errors`` is set to ``"filter"``, then in case of errors less than
        n_samples values may be returned. If ``errors`` is set to ``"ignore"``,
        some molecules may be returned without conformers generated and with
        ``conf_id`` property set to -1.

        Parameters
        ----------
        X : {sequence, array-like} of shape (n_samples,)
            Sequence containing RDKit ``Mol`` objects.

        y : np.ndarray of shape (n_samples,)
            Array with labels for molecules.

        copy : bool, default=True
            Copy the input X or not. In contrast to most classes, input molecules
            are copied by default, since RDKit modifies them with conformers in place.

        Returns
        -------
        X : list of shape (n_samples_conf_gen,)
            List with RDKit ``PropertyMol`` objects, each one with conformers computed
            and ``conf_id`` integer property set.

        y : np.ndarray of shape (n_samples_conf_gen,)
            Array with labels for molecules.
        """
        return self._transform(X, y, copy)

    def transform(self, X: Sequence[Mol], copy: bool = False) -> list[PropertyMol]:
        """
        Generate conformers for molecules.

        If ``errors`` is set to ``"filter"``, then in case of errors less than
        n_samples values may be returned. If ``errors`` is set to ``"ignore"``,
        some molecules may be returned without conformers generated and with
        ``conf_id`` property set to -1.

        Parameters
        ----------
        X : {sequence, array-like} of shape (n_samples,)
            Sequence containing RDKit ``Mol`` objects.

        copy : bool, default=True
            Copy the input X or not. In contrast to most classes, input molecules
            are copied by default, since RDKit modifies them with conformers in place.

        Returns
        -------
        X : list of shape (n_samples_conf_gen,)
            List with RDKit ``PropertyMol`` objects, each one with conformers computed
            and ``conf_id`` integer property set.
        """
        y = np.empty(len(X))
        X, y = self._transform(X, y, copy)
        return X

    def _transform(
        self, X: Sequence[Mol], y: np.ndarray, copy: bool = True
    ) -> tuple[list[PropertyMol], np.ndarray]:
        mols = super().transform(X)

        if copy:
            y = deepcopy(y)

        # keep only molecules and labels for which we generated conformers
        mols_with_conformers = []
        idxs_to_keep = []
        for idx, mol in enumerate(mols):
            if mol.GetIntProp("conf_id") != -1 or self.errors == "ignore":
                mols_with_conformers.append(mol)
                idxs_to_keep.append(idx)

        y = y[idxs_to_keep]

        return mols_with_conformers, y

    def _transform_batch(self, X: Sequence[Mol]) -> list[Mol]:
        # adding hydrogens and sanitizing is recommended for conformer generation
        mols = [AddHs(mol) for mol in X]
        for mol in mols:
            SanitizeMol(mol)

        # make sure properties are pickled properly
        mols = [PropertyMol(mol) for mol in mols]

        mols_and_conf_ids = [self._embed_molecule(mol) for mol in mols]

        if self.num_conformers > 1:
            mols_and_conf_ids = [
                (mol, self._select_conformer(mol)) for mol, conf_id in mols_and_conf_ids
            ]

        mols = []
        for mol, conf_id in mols_and_conf_ids:
            mol.SetIntProp("conf_id", conf_id)
            mols.append(mol)

        return mols

    def _embed_molecule(self, mol: Mol) -> tuple[Mol, int]:
        # we create a new embedding params for each molecule, since it can
        # get modified if default settings fail to generate conformers
        embed_params = ETKDGv3()
        embed_params.useSmallRingTorsions = True
        embed_params.trackFailures = True
        embed_params.randomSeed = self.random_state

        # basic attempt
        if self.num_conformers == 1:
            embedder = EmbedMolecule
        else:
            embedder = partial(EmbedMultipleConfs, numConfs=self.num_conformers)

        conf_id = embedder(mol, params=embed_params)

        if conf_id == -1:
            # more tries
            embed_params.maxIterations = self.max_gen_attempts
            embed_params.useRandomCoords = True
            conf_id = embedder(mol, params=embed_params)

        if conf_id == -1:
            # even more tries, turn off conditions
            embed_params.maxIterations = 10 * self.max_gen_attempts
            embed_params.enforceChirality = False
            embed_params.ignoreSmoothingFailures = True
            conf_id = embedder(mol, params=embed_params)

        # we should not fail at this point
        if conf_id == -1:
            smiles = MolToSmiles(RemoveHs(mol))
            fail_reason = self._print_conf_gen_failure_reason(embed_params)
            if self.errors == "raise":
                raise ValueError(
                    f"Could not generate conformer for {smiles}:\n{fail_reason}"
                )
            elif self.verbose:
                print(f"Could not generate conformer for {smiles}:\n{fail_reason}")
                return mol, -1

        if self.optimize_force_field:
            self._optimize_conformers(mol)

        return mol, conf_id

    def _print_conf_gen_failure_reason(self, embed_params: ETKDGv3) -> str:
        fail_idx_to_name = {idx: name for name, idx in EmbedFailureCauses.names.items()}
        fail_counts = embed_params.GetFailureCounts()
        fail_names_with_counts = [
            f"{fail_idx_to_name[idx]}: {fail_counts[idx]}"
            for idx in range(len(fail_counts))
        ]
        fail_reason = "\n".join(fail_names_with_counts)
        return fail_reason

    def _optimize_conformers(self, mol: Mol) -> None:
        for conf in mol.GetConformers():
            ff = self._get_force_field(mol, conf_id=conf.GetId())
            ff.Minimize()

    def _select_conformer(self, mol: Mol) -> int:
        if self.multiple_confs_select == "first":
            return next(mol.GetConformers())
        else:  # min_energy
            energies = np.empty((mol.GetNumConformers(),))
            for i, conf in enumerate(mol.GetConformers()):
                ff = self._get_force_field(mol, conf_id=conf.GetId())
                energies[i] = ff.CalcEnergy()
            return int(np.argmin(energies))

    def _get_force_field(self, mol: Mol, conf_id: int) -> ForceField:
        if self.optimize_force_field == "UFF":
            return UFFGetMoleculeForceField(mol, confId=conf_id)
        else:
            # MMFF94 or MMFF94s
            MMFFSanitizeMolecule(mol)
            mmff_props = MMFFGetMoleculeProperties(
                mol, mmffVariant=self.optimize_force_field
            )
            return MMFFGetMoleculeForceField(mol, mmff_props, confId=conf_id)
